from typing import Dict, List
from .ContentDataclasses import ContentDocument, DocumentContext, DocumentUnit
from lxml import etree

XHTML_NAMESPACE: str = "http://www.w3.org/1999/xhtml"
XML_NAMESPACE: str = "http://www.w3.org/XML/1998/namespace"
IXBRL_NAMESPACE: str = "http://www.xbrl.org/2013/inlineXBRL"
LINK_NAMESPACE: str = "http://www.xbrl.org/2003/linkbase" 
XLINK_NAMESPACE: str = "http://www.w3.org/1999/xlink"
INSTANCE_NAMESPACE: str = "http://www.xbrl.org/2003/instance"
DIMENSIONS_NAMESPACE: str = "http://xbrl.org/2006/xbrldi"
XSI_NAMESPACE: str = "http://www.w3.org/2001/XMLSchema-instance"

class BaseProducer:
    content_document: ContentDocument
    tag_id_tracker: Dict[str, etree.Element]
    local_namespace: str
    local_namespace_prefix: str
    local_taxonomy_schema: str

    contexts: Dict[str, 'DocumentContext']
    units: Dict[str, 'DocumentUnit']

    def __init__(cls, document: ContentDocument, local_namespace: str = None, local_namespace_prefix: str = None, local_taxonomy_schema: str = None):
        cls.content_document = document
        cls.tag_id_tracker = {}
        cls.local_namespace = local_namespace
        cls.local_namespace_prefix = local_namespace_prefix
        cls.local_taxonomy_schema = local_taxonomy_schema

        cls.contexts = {}
        cls.units = {}

        for tag in cls.content_document._get_applied_tags():
            # check context
            tag_context: DocumentContext = tag.to_document_context()
            found_context_id: str = None
            for known_context_id, known_context in cls.contexts.items():
                if known_context.euqals(tag_context):
                    found_context_id = known_context_id
                    break
            if not found_context_id:
                found_context_id: str = f"c-{len(cls.contexts.keys())}"
                cls.contexts[found_context_id] = tag_context
            tag.context_id = found_context_id
            # check unit
            if tag.attributes.unit:
                tag_unit: DocumentUnit = tag.attributes.unit
                found_unit_id: str = None
                for known_unit_id, known_unit in cls.units.items():
                    if known_unit.equals(tag_unit):
                        found_unit_id = known_unit_id
                        break
                if not found_unit_id:
                    found_unit_id: str = f"u-{len(cls.units.keys())}"
                    cls.units[found_unit_id] = tag_unit
                tag.attributes.unit_ref = found_unit_id

    def _add_context_elements(cls, parent_element: etree._Element) -> None:
        for context_id, context in cls.contexts.items():
            context_element: etree._Element = etree.SubElement(parent_element, f"{{{INSTANCE_NAMESPACE}}}context", {"id":context_id})
            entity_element: etree._Element = etree.SubElement(context_element, f"{{{INSTANCE_NAMESPACE}}}entity")
            entity_identifier_element: etree._Element = etree.SubElement(
                entity_element, 
                f"{{{INSTANCE_NAMESPACE}}}identifier",
                {
                    "scheme": context.entity_scheme
                }
            )
            entity_identifier_element.text = context.entity
            period_element: etree._Element = etree.SubElement(context_element, f"{{{INSTANCE_NAMESPACE}}}period")
            if context.start_date:
                period_start_element: etree._Element = etree.SubElement(period_element, f"{{{INSTANCE_NAMESPACE}}}startDate")
                period_start_element.text = context.start_date
                period_end_element: etree._Element = etree.SubElement(period_element, f"{{{INSTANCE_NAMESPACE}}}endDate")
                period_end_element.text = context.end_date
            else:
                period_instant_element: etree._Element = etree.SubElement(period_element, f"{{{INSTANCE_NAMESPACE}}}instant")
                period_instant_element.text = context.end_date
            if len(context.dimensions):
                scenario_element: etree._Element = etree.SubElement(context_element, f"{{{INSTANCE_NAMESPACE}}}scenario")
                for dimension in context.dimensions:
                    if dimension.typed_member_value == None:
                        explicit_dimension_element: etree._Element = etree.SubElement(
                            scenario_element, 
                            f"{{{DIMENSIONS_NAMESPACE}}}explicitMember",
                            {
                                "dimension": dimension.axis.to_prefixed_name(
                                    cls.content_document.namespaces, 
                                    cls.local_namespace_prefix
                                )
                            }
                        )
                        explicit_dimension_element.text = dimension.member.to_prefixed_name(
                            cls.content_document.namespaces, 
                            cls.local_namespace_prefix
                        )
                    else:
                        typed_dimension_element: etree._Element = etree.SubElement(
                            scenario_element, 
                            f"{{{DIMENSIONS_NAMESPACE}}}typedMember",
                            {
                                "dimension": dimension.axis.to_prefixed_name(cls.content_document.namespaces, cls.local_namespace_prefix)
                            }
                        )
                        typed_member_element: etree._Element = etree.SubElement(
                            typed_dimension_element,
                            f"{{{dimension.member.namespace}}}{dimension.member.name}"
                        )
                        typed_member_element.text = dimension.typed_member_value

    def _add_unit_elements(cls, parent_element: etree._Element) -> None:
        for unit_id, unit in cls.units.items():
            unit_element: etree._Element = etree.SubElement(
                parent_element, 
                f"{{{INSTANCE_NAMESPACE}}}unit", 
                {"id": unit_id}
            )
            if unit.denominator:
                divide_element: etree._Element = etree.SubElement(unit_element, f"{{{INSTANCE_NAMESPACE}}}divide")
                numerator_element: etree._Element = etree.SubElement(divide_element, f"{{{INSTANCE_NAMESPACE}}}unitNumerator")
                numerator_measure_element: etree._Element = etree.SubElement(numerator_element, f"{{{INSTANCE_NAMESPACE}}}measure")
                numerator_measure_element.text = unit.numerator.to_prefixed_name(cls.content_document.namespaces)
                denominator_element: etree._Element = etree.SubElement(divide_element, f"{{{INSTANCE_NAMESPACE}}}unitDenominator")
                denominator_measure_element: etree._Element = etree.SubElement(denominator_element, f"{{{INSTANCE_NAMESPACE}}}measure")
                denominator_measure_element.text = unit.denominator.to_prefixed_name(cls.content_document.namespaces)
            else:
                measure_element: etree._Element = etree.SubElement(unit_element, f"{{{INSTANCE_NAMESPACE}}}measure")
                measure_element.text = unit.numerator.to_prefixed_name(cls.content_document.namespaces)



from typing import List
import os
import logging

from .logger_setup import logger_conf
logger_conf()

from .xbrl_generation.ContentDataclasses import ContentDocument

from .xbrl_generation.XbrlProducer import XbrlProducer
from .xbrl_generation.PackageDataclasses import File
from .xbrl_generation.InputData import InputData
from .xbrl_generation.HtmlProducer import HtmlProducer
from .xbrl_generation.TaxonomyProducer import TaxonomyProducer

from .utils.schema_validation import validate_schema

from .file_conversion import doc_to_data

logger = logging.getLogger(__name__)

SCHEMA_FOLDER: str = os.path.join(os.path.dirname(os.path.realpath(__file__)), "schemas")

def convert_document(document_path: str) -> InputData:
    logger.info(f"Converting file {document_path}")
    input_data_object = doc_to_data(document_path)
    validate_input_data(input_data_object.to_dict())
    return input_data_object

def validate_input_data(data: dict) -> None:
    logger.info(f"Validating input data")
    # get schemas
    input_schema_folder = os.path.join(SCHEMA_FOLDER, "input")
    validate_schema(data, "https://xbrl-forge.org/schema/input/wrapper", input_schema_folder)

def load_input_data(data: dict) -> InputData:
    logger.info(f"Loading Input Data")
    return InputData.from_dict(data)

def create_xbrl(input_data_list: List[InputData], xthml_template: str = None) -> File:
    logger.info(f"Creating XBRL")
    # load data
    loaded_data: InputData = InputData.combine(input_data_list)
    local_namespace = None
    local_namespace_prefix = None
    local_taxonomy_schema = None
    if loaded_data.taxonomy:
        local_namespace=loaded_data.taxonomy.namespace
        local_namespace_prefix=loaded_data.taxonomy.prefix 
        local_taxonomy_schema=loaded_data.taxonomy.schema_url
    reports_folder: File = None
    untagged_reports_folder: File = None
    inline_instances: int = 0
    non_inline_instances: int = 0
    for report in loaded_data.reports:
        if not reports_folder:        
            reports_folder = File("reports", contained_files=[])
        if report.xhtml:
            html_producer: HtmlProducer = HtmlProducer(
                report, 
                xthml_template=xthml_template, 
                local_namespace=local_namespace, 
                local_namespace_prefix=local_namespace_prefix, 
                local_taxonomy_schema=local_taxonomy_schema
            )
            if html_producer.ixbrl:
                reports_folder.contained_files.append(html_producer.create_html())
                inline_instances += 1
            else:
                if not untagged_reports_folder:
                    untagged_reports_folder: File = File("untagged_reports", contained_files=[])
                    reports_folder.contained_files.append(untagged_reports_folder)
                untagged_reports_folder.contained_files.append(html_producer.create_html())
        else:
            xbrl_producer: XbrlProducer = XbrlProducer(
                report, 
                local_namespace=local_namespace, 
                local_namespace_prefix=local_namespace_prefix, 
                local_taxonomy_schema=local_taxonomy_schema
            )
            reports_folder.contained_files.append(xbrl_producer.create_xbrl())
            non_inline_instances += 1
    if not loaded_data.taxonomy:
        return reports_folder
    taxonomy_producer: TaxonomyProducer = TaxonomyProducer(loaded_data.taxonomy)
    package_extension: str = "zip"
    if inline_instances == 1 and non_inline_instances == 0: package_extension = "xbri"
    if inline_instances == 0 and non_inline_instances == 1: package_extension = "xbr"
    return taxonomy_producer.create_files(reports_folder, package_extension)


from dataclasses import dataclass
from typing import Dict, List, Tuple
import os
import shutil
import logging
import zipfile

logger = logging.getLogger(__name__)

class File:
    name: str
    content: str 
    contained_files: List['File'] 
    zip_extension: str

    def __init__(cls, name: str, content: str = None, contained_files: List['File'] = None, zip_extension: str = "zip"):
        cls.name = name
        cls.content = content
        cls.contained_files = [] if contained_files == None else contained_files
        cls.zip_extension = zip_extension

    def save_files(cls, folder_path: str, remove_existing_files: bool = False) -> None:
        new_path: str = os.path.join(folder_path, cls.name)
        if remove_existing_files:
            if os.path.isdir(new_path):
                shutil.rmtree(new_path)
            if os.path.isfile(new_path):
                os.remove(new_path)
        if cls.contained_files:
            os.mkdir(new_path)
            for file in cls.contained_files:
                file.save_files(new_path, remove_existing_files)
        else:
            with open(new_path, "w+") as f:
                f.write(cls.content)

    def create_package(cls, folder_path: str, remove_existing_package: bool = False) -> str:
        file_path: str = os.path.join(folder_path, f"{cls.name}.{cls.zip_extension}")
        if remove_existing_package:
            if os.path.isfile(file_path):
                os.remove(file_path)
        with zipfile.ZipFile(file_path, "w") as zip:
            for path, file in cls._list_files():
                zip.writestr(path, file.content)
        return file_path

    def _list_files(cls, prepend_path: str = None) -> List[Tuple[str, "File"]]:
        file_path: str = cls.name
        if prepend_path:
            file_path = os.path.join(prepend_path, file_path)
        file_list: List[Tuple[str, "File"]] = []
        if cls.content:
            file_list.append((file_path, cls))
        for child in cls.contained_files:
            file_list += child._list_files(file_path)
        return file_list

@dataclass
class Tag:
    namespace: str
    name: str

    @classmethod
    def from_dict(cls, data: dict) -> 'Tag':
        return cls(
            namespace=data.get("namespace"),
            name=data.get("name")
        )
    
    def copy(cls) -> 'Tag':
        return cls.__class__(
            namespace=cls.namespace,
            name=cls.name
        )
    
    def to_uname(cls) -> str:
        return f"{{{cls.namespace}}}{cls.name}"

    def to_prefixed_name(cls, prefixes: Dict[str, str], local_taxonomy_prefix: str = None) -> str:
        if not cls.namespace:
            return f"{local_taxonomy_prefix}:{cls.name}"
        return f"{prefixes.get(cls.namespace, 'unknown')}:{cls.name}"
    
    def to_dict(cls) -> dict:
        return {
            "namespace": cls.namespace,
            "name": cls.name
        }



import json

from src.xbrl_forge import create_xbrl, validate_input_data, load_input_data, convert_document

# example xhtml template
with open("examples/input_documents/xhtml_template.html", "r") as f:
    xthml_template = f.read()

# example esef file
with open("examples/input_documents/ESEF_example.json", "r") as f:
    esef_json = json.loads(f.read())
    validate_input_data(esef_json)
    esef_loaded = load_input_data(esef_json)

# example combination of different document types
with open("examples/input_documents/ixbrl_xhtml_example.json", "r") as f:
    ixbrl_xhtml_json = json.loads(f.read())
    validate_input_data(ixbrl_xhtml_json)
    ixbrl_xhtml_loaded = load_input_data(ixbrl_xhtml_json)
with open("examples/input_documents/xbrl_example.json", "r") as f:
    xbrl_json = json.loads(f.read())
    validate_input_data(xbrl_json)
    xbrl_loaded = load_input_data(xbrl_json)
loaded_docx_xhtml = convert_document("examples/file_conversions/Testing Docx document.docx")
    
# create example esef file
results_esef = create_xbrl([esef_loaded], xthml_template=xthml_template)
results_esef.save_files("examples/result", True)
results_esef.create_package("examples/result", True)

# create example xbrl file
results_esef = create_xbrl([xbrl_loaded])
results_esef.save_files("examples/result", True)
results_esef.create_package("examples/result", True)

# create example combined report
results = create_xbrl([esef_loaded, ixbrl_xhtml_loaded, xbrl_loaded, loaded_docx_xhtml], xthml_template=xthml_template)
results.save_files("examples/result", True)
results.create_package("examples/result", True)


import logging
import os

def logger_conf(log_level: int = None):
    # if not provided directly
    if not log_level:
        # check env, defaut to info
        log_level = logging.getLevelNamesMapping().get(os.environ.get("XBRL_FORGE_LOGGING", "WARNING"))
    logging.basicConfig(level=log_level, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger(__name__)
    logger.info(f"Configured Logger to level {logging.getLevelName(log_level)}")


from typing import Dict, List
import mammoth
from lxml import etree
import re

from ..xbrl_generation.ContentDataclasses import ContentItem, ListElement, ListItem, TitleItem, ParagraphItem, ImageItem, TableRow, TableCell, TableItem
from .BaseLoader import BaseLoader

DATA_WRAPPER_NAME: str = "dataWrapper"

class DocxLoader(BaseLoader):

    def __init__(cls):
        super(DocxLoader, cls).__init__()

    def load_document(cls, path: str) -> List[ContentItem]:
        with open(path, "rb") as docx_content:
            result = mammoth.convert_to_html(docx_content)
        data = etree.fromstring(f"<{DATA_WRAPPER_NAME}>{result.value}</{DATA_WRAPPER_NAME}>")
        cls.content = cls._add_to_content(data)
        return cls.content

    def _add_to_content(cls, element: etree._Element) -> List[ContentItem]:
        content: List[ContentItem] = []
        if element.tag == DATA_WRAPPER_NAME:
            for child in element.getchildren():
                content += cls._add_to_content(child)
            return content
        # add paragraph (and image)
        if element.tag == "p":
            # if there is one child image, only add that one
            children: List[etree._Element] = element.getchildren()
            if children and children[0].tag == "img":
                image_element = children[0]
                content.append(ImageItem(
                    type=ContentItem.TYPE_IMAGE,
                    tags=[],
                    image_data=image_element.attrib.get("src", "")
                ))
            else:
                content.append(ParagraphItem(
                    type=ContentItem.TYPE_PARAGRAPH,
                    tags=[],
                    content=element.text
                ))
        # add title
        elif re.match(r"h[0-9]", element.tag):
            content.append(TitleItem(
                type=ContentItem.TYPE_TITLE,
                tags=[],
                content=element.text,
                level=int(element.tag[1]) 
            ))
        # add lists
        elif element.tag in ["ol", "ul"]:
            list_data: List[ListElement] = []
            list_children: List[etree._Element] = element.getchildren()
            for list_child_element in list_children:
                list_data_elemet: ListElement = ListElement([])
                if list_child_element.text:
                    list_data_elemet.content.append(ParagraphItem(
                        type=ContentItem.TYPE_PARAGRAPH,
                        tags=[],
                        content=list_child_element.text
                    ))
                for list_child_sub_child in list_child_element.getchildren():
                    list_data_elemet.content += cls._add_to_content(list_child_sub_child)
                list_data.append(list_data_elemet)
            content.append(ListItem(
                type=ContentItem.TYPE_LIST,
                tags=[],
                elements=list_data,
                ordered=element.tag[0] == "o"
            ))
        # add tables
        elif element.tag == "table":
            table_rows: List[TableRow] = []
            row_element: etree._Element
            for row_element in element.getchildren():
                row_data = TableRow([])
                cell_element: etree.Element
                for cell_element in row_element.getchildren():
                    cell_content: List[ContentItem] = []
                    if cell_element.text:
                        cell_content.append(ParagraphItem(
                            type=ContentItem.TYPE_PARAGRAPH,
                            tags=[],
                            content=cell_element.text
                        ))
                    sub_element: etree._Element
                    for sub_element in cell_element.getchildren():
                        cell_content += cls._add_to_content(sub_element)
                    row_data.cells.append(TableCell(
                        content=cell_content,
                        header=False,
                        rowspan=int(cell_element.attrib.get("rowspan", 1)),
                        colspan=int(cell_element.attrib.get("colspan", 1))
                    ))
                table_rows.append(row_data)
            content.append(TableItem(
                type=ContentItem.TYPE_TABLE,
                tags=[],
                rows=table_rows
            ))
        else:
            raise Exception(f"Unknown Tag: {element.tag}")
        return content


from typing import List

from ..xbrl_generation.ContentDataclasses import ContentItem

class BaseLoader():
    content: List[ContentItem]

    def __init__(cls):
        cls.content = []

    def load_document(cls, path: str) -> List[ContentItem]:
        return cls.content


from typing import Dict, Tuple
from lxml import etree
import logging


from .ElementRender import render_content

from .PackageDataclasses import File
from .ContentDataclasses import AppliedTag, AppliedTagTree, ContentDocument, ContentItem
from .BaseProducer import BaseProducer, XHTML_NAMESPACE, XML_NAMESPACE, IXBRL_NAMESPACE, LINK_NAMESPACE, XLINK_NAMESPACE, INSTANCE_NAMESPACE, DIMENSIONS_NAMESPACE, XSI_NAMESPACE
from .utils import xml_to_string
from .HtmlTemplate import get_xhtml_template

logger = logging.getLogger(__name__)

class HtmlProducer(BaseProducer):
    ixbrl: bool
    xhtml_template: str

    def __init__(cls, document: ContentDocument, xthml_template: str = "", local_namespace: str = None, local_namespace_prefix: str = None, local_taxonomy_schema: str = None):
        super().__init__(
            document=document, 
            local_namespace=local_namespace, 
            local_namespace_prefix=local_namespace_prefix,
            local_taxonomy_schema=local_taxonomy_schema
        )
        cls.ixbrl = len(cls.contexts) > 0
        cls.xhtml_template = xthml_template

    def create_html(cls) -> File:
        # Populate Namespaces
        namespace_map = {
            None: XHTML_NAMESPACE,
            "xml": XML_NAMESPACE,
        }

        if cls.ixbrl:
            namespace_map.update({
                "ix": IXBRL_NAMESPACE,
                "link": LINK_NAMESPACE,
                "xlink": XLINK_NAMESPACE,
                "xbrli": INSTANCE_NAMESPACE,
                "xbrldi": DIMENSIONS_NAMESPACE,
                "xsi": XSI_NAMESPACE
            })
            if cls.local_namespace:
                namespace_map[cls.local_namespace_prefix] = cls.local_namespace
            for namespace, prefix in cls.content_document.namespaces.items():
                namespace_map[prefix] = namespace

        # Create basic XHTML strucure
        xhtml_root, xhtml_body, xhtml_content_root = get_xhtml_template(
            name = cls.content_document.name,
            namespace_map = namespace_map,
            xhtml_template = cls.xhtml_template
        )

        if cls.ixbrl:
            # create ixbrl header information
            ixbrl_header_container: etree._Element = etree.SubElement(xhtml_body, f"{{{XHTML_NAMESPACE}}}div", {"style":"display:none;"})
            cls.ixbrl_header: etree._Element = etree.SubElement(ixbrl_header_container, f"{{{IXBRL_NAMESPACE}}}header")
            cls.ixbrl_hidden: etree._Element = None
            ixbrl_references: etree._Element = etree.SubElement(cls.ixbrl_header, f"{{{IXBRL_NAMESPACE}}}references", {f"{{{XML_NAMESPACE}}}lang": cls.content_document.lang})
            cls.schema_url = cls.content_document.taxonomy_schema if cls.content_document.taxonomy_schema else cls.local_taxonomy_schema
            schema_ref: etree._Element = etree.SubElement(
                ixbrl_references, 
                f"{{{LINK_NAMESPACE}}}schemaRef",
                {
                    f"{{{XLINK_NAMESPACE}}}href": cls.schema_url,
                    f"{{{XLINK_NAMESPACE}}}type": "simple"
                }
            )
            ixbrl_resources: etree._Element = etree.SubElement(cls.ixbrl_header, f"{{{IXBRL_NAMESPACE}}}resources")
        
            # add contexts to header
            cls._add_context_elements(ixbrl_resources)
            
            # Add Units
            cls._add_unit_elements(ixbrl_resources)
            
        # Add html contents
        for content in cls.content_document.content:
            cls._convert_element(content, xhtml_content_root)

        return File(name=f"{cls.content_document.name}.html", content=xml_to_string(xhtml_root))
    
    def _convert_element(cls, content_item: ContentItem, parent: etree.Element) -> None:
        # check if tags are applied to the whole structure
        complete_tags = [tag for tag in content_item.tags if not tag.end_index and not tag.start_index]
        for tag in complete_tags:
            parent, new_element = cls._create_ixbrl_tag(tag, parent)
            parent = new_element
        # prepare part tags for the application to the text
        part_tags = [tag for tag in content_item.tags if tag.end_index or tag.start_index]
        # add content based on type
        render_content(
            content_item,
            parent,
            cls._add_text_with_tags_to_element,
            cls._convert_element,
            part_tags
        )

    def _create_ixbrl_tag(cls, tag: AppliedTag, parent: etree.Element) -> Tuple[etree.Element, etree.Element]:
        prefixed_name: str = tag.to_prefixed_name(cls.content_document.namespaces, cls.local_namespace_prefix)
        tag_id_base = f'{prefixed_name.replace(":", "_")}_{tag.context_id}_-_'
        id_number: int = 0
        # get previous is if known
        previous_element: etree._Element = None
        if tag_id_base in cls.tag_id_tracker:
            previous_element = cls.tag_id_tracker[tag_id_base]
            id_number = int(previous_element.attrib["id"].split("_")[-1]) + 1
        # add tag
        # if the tag attributes contain a unit, then it must be a numeric tag
        if tag.attributes.unit_ref:
            new_element: etree._Element = etree.SubElement(
                parent,
                f"{{{IXBRL_NAMESPACE}}}nonFraction",
                {
                    "id": f"{tag_id_base}{id_number}",
                    "name": prefixed_name,
                    "contextRef": tag.context_id,
                    "unitRef": tag.attributes.unit_ref,
                    "scale": str(tag.attributes.scale),
                    "decimals": str(tag.attributes.decimals)
                }
            )
            if tag.attributes.sign:
                new_element.attrib["sign"] = "-"
            if tag.attributes.format:
                new_element.attrib["format"] = tag.attributes.format.to_prefixed_name(cls.content_document.namespaces)
            if tag.attributes.nil:
                new_element.attrib[f"{{{XSI_NAMESPACE}}}nil"] = "true"
        # otherwise it must be a non-numeric tag
        else:
            # if its not an enum, it can be tagged inline
            if not tag.attributes.enumeration_values:
                if previous_element == None:
                    new_element = etree.SubElement(
                        parent,
                        f"{{{IXBRL_NAMESPACE}}}nonNumeric",
                        {
                            "id": f"{tag_id_base}{id_number}",
                            "name": prefixed_name,
                            "contextRef": tag.context_id
                        }
                    )
                    if tag.attributes.escape:
                        new_element.attrib["escape"] = "true"
                    if tag.attributes.format:
                        new_element.attrib["format"] = tag.attributes.format.to_prefixed_name(cls.content_document.namespaces)
                    if tag.attributes.nil:
                        new_element.attrib[f"{{{XSI_NAMESPACE}}}nil"] = "true"
                else:
                    previous_element.attrib["continuedAt"] = f"{tag_id_base}{id_number}"
                    new_element = etree.SubElement(
                        parent,
                        f"{{{IXBRL_NAMESPACE}}}continuation",
                        {
                            "id": f"{tag_id_base}{id_number}"
                        }
                    )
            # enumerations
            else:
                # check if the hidden element already exists
                if cls.ixbrl_hidden == None:
                    cls.ixbrl_hidden = etree.SubElement(
                        cls.ixbrl_header,
                        f"{{{IXBRL_NAMESPACE}}}hidden"
                    )
                hidden_tag_element: etree._Element = etree.SubElement(
                    cls.ixbrl_hidden,
                    f"{{{IXBRL_NAMESPACE}}}nonNumeric",
                    {
                        "id": f"{tag_id_base}{id_number}",
                        "name": prefixed_name,
                        "contextRef": tag.context_id
                    }
                )
                # add enum values
                hidden_tag_element.text = " ".join([enum.value(cls.local_namespace) for enum in tag.attributes.enumeration_values])
                new_element = parent
        # add element to known ids
        cls.tag_id_tracker[tag_id_base] = new_element
        return parent, new_element
    
    def _add_text_with_tags_to_element(cls, element: etree.Element, tag_tree: AppliedTagTree, content: str) -> None:
        start_text: str = content[tag_tree.item.start_index:tag_tree.item.end_index]
        if tag_tree.children:
            start_text = content[tag_tree.item.start_index:tag_tree.children[0].item.start_index]
        element.text = start_text
        for child_index, child_tree in enumerate(tag_tree.children):
            current_element, child_element = cls._create_ixbrl_tag(child_tree.item, element)
            cls._add_text_with_tags_to_element(child_element, child_tree, content)
            # if it is not the last element
            if child_index < len(tag_tree.children) - 1:
                child_element.tail = content[child_tree.item.end_index:tag_tree.children[child_index + 1].item.start_index]
            else:
                child_element.tail = content[child_tree.item.end_index:tag_tree.item.end_index]



from typing import Dict, Tuple
from lxml import etree
from .BaseProducer import XHTML_NAMESPACE
from copy import deepcopy
import logging

logger = logging.getLogger(__name__)

XHTML_CONTENT_ROOT_ID: str = "xhtml-content-root"

DEFAULT_TEMPLATE: str = """
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
    <title>DEFAULT_TEMPLATE</title>
    <style type="text/css">
        body { background-color: gray; padding-left: 10px; padding-right: 10px; padding-top: 0px; padding-bottom: 0px; font-size: 16px; }
        #xhtml-content-root { background-color: white; box-shadow: 5px 5px 5px black; padding: 10px; width: 90%; }
        h1 { border-bottom: 1px solid black; font-size: 20px; }
        h2, h3, h4, h5, h6 {}
        p {}
        table, th, td { border: 1px solid black; border-collapse: collapse; padding: 5px; }
        img { width: 100%; height: auto; }
        ul {}
        ol { list-style-type: lower-alpha; }
        li {}
    </style>
</head>
<body>
    <div id="xhtml-content-root"></div>
</body>
</html>
"""

def get_xhtml_template(name: str, namespace_map: Dict[str, str], xhtml_template: str) -> Tuple[etree._Element, etree._Element, etree._Element]:
    # try to use provided template, else use default
    if xhtml_template:
        logger.debug("Trying to load custom XHTML Template...")
        try:
            return _load_template(name, namespace_map, xhtml_template)
        except Exception as e:
            logger.error(f"Could not use provided custom XHTML Template: {str(e)}")
    logger.debug("Using default XHTML Template...")
    return _load_template(name, namespace_map, DEFAULT_TEMPLATE)

def _load_template(name: str, namespace_map: Dict[str, str], xhtml_template: str) -> Tuple[etree._Element, etree._Element, etree._Element]:
    template_root: etree._Element = etree.fromstring(xhtml_template)
    xhtml_root: etree._Element = etree.Element(f"{{{XHTML_NAMESPACE}}}html", nsmap=namespace_map)
    child: etree._Element
    for child in template_root:
        xhtml_root.append(child)
    xhtml_body: etree._Element = xhtml_root.find("body", { None: XHTML_NAMESPACE })
    xhtml_title: etree._Element = xhtml_root.find(".//title", { None: XHTML_NAMESPACE })
    xhtml_title.text = name
    xhtml_content_root: etree._Element = xhtml_root.find(".//div[@id='xhtml-content-root']", { None: XHTML_NAMESPACE })
    return xhtml_root, xhtml_body, xhtml_content_root



from typing import Callable, Dict, List
from venv import logger
from .ContentDataclasses import AppliedTag, AppliedTagTree, BaseXbrlItem, ContentItem, ImageItem, ListItem, ParagraphItem, TableItem, TitleItem
from lxml import etree

XHTML_NAMESPACE: str = "http://www.w3.org/1999/xhtml"

def render_content(content_item: ContentItem, parent: etree.Element, add_text: Callable[[etree.Element, AppliedTagTree, str], None], recusive: Callable[[ContentItem, etree.Element], None], part_tags: List[AppliedTag] = []) -> None:
    match content_item.type:
        case ContentItem.TYPE_TITLE:
            title_item: TitleItem = content_item
            # levels 1 - 6 are available
            level = title_item.level
            if level < 1: level = 1
            if level > 6: level = 6
            heading: etree.Element = etree.SubElement(parent, f"{{{XHTML_NAMESPACE}}}h{level}")
            tag_tree: AppliedTagTree = AppliedTag.create_tree(part_tags, len(title_item.content))
            add_text(heading, tag_tree, title_item.content)
        case ContentItem.TYPE_PARAGRAPH:
            paragraph_item: ParagraphItem = content_item
            paragraph_element: etree.Element = etree.SubElement(parent, f"{{{XHTML_NAMESPACE}}}p")
            tag_tree: AppliedTagTree = AppliedTag.create_tree(part_tags, len(paragraph_item.content))
            add_text(paragraph_element, tag_tree, paragraph_item.content)
        case ContentItem.TYPE_TABLE:
            table_item: TableItem = content_item
            # part tags are ignored for this item, a table can only be tagged as a whole or individual cells
            table_element: etree.Element = etree.SubElement(parent, f"{{{XHTML_NAMESPACE}}}table")
            # create a row for every row in the item
            for row_item in table_item.rows:
                row_element: etree.Element = etree.SubElement(table_element, f"{{{XHTML_NAMESPACE}}}tr")
                # create a cell for every cell element, depending on header or not
                for cell_item in row_item.cells:
                    attributes: Dict[str, str] = {}
                    if cell_item.colspan > 1:
                        attributes["colspan"] = str(cell_item.colspan)
                    if cell_item.rowspan > 1:
                        attributes["rowspan"] = str(cell_item.rowspan)
                    if cell_item.header:
                        cell_element: etree.Element = etree.SubElement(row_element, f"{{{XHTML_NAMESPACE}}}th", attributes)
                    else:
                        cell_element: etree.Element = etree.SubElement(row_element, f"{{{XHTML_NAMESPACE}}}td", attributes)
                    # add content to cell
                    for cell_content_item in cell_item.content:
                        recusive(cell_content_item, cell_element)                    
        case ContentItem.TYPE_IMAGE:
            image_item: ImageItem = content_item
            # part tags are ignored for this item, a image can only be tagged as a whole
            image_element: etree.Element = etree.SubElement(
                parent, 
                f"{{{XHTML_NAMESPACE}}}img",
                {
                    "src": image_item.image_data,
                    "alt": "image"
                }
            )
        case ContentItem.TYPE_LIST:
            list_item: ListItem = content_item
            # part tags are ignored for this item, a image can only be tagged as a whole or the content of each list element
            # the container element depends on ordered or unordered
            if list_item.ordered:
                list_element: etree.Element = etree.SubElement(parent, f"{{{XHTML_NAMESPACE}}}ol")
            else:
                list_element: etree.Element = etree.SubElement(parent, f"{{{XHTML_NAMESPACE}}}ul")
            # add list items to the list element
            for list_content_item in list_item.elements:
                list_content_element: etree.Element = etree.SubElement(list_element, f"{{{XHTML_NAMESPACE}}}li")
                # add the content of the element to the list item elemnet
                for list_element_content_item in list_content_item.content:
                    recusive(list_element_content_item, list_content_element)
        case ContentItem.TYPE_BASE_XBRL:
            text_item: BaseXbrlItem = content_item
            if not parent.text:
                parent.text = ""
            parent.text += text_item.content
        case _:
            logger.error(f"Could not convert element of type '{content_item.type}' to XHTML.")


from typing import Dict, List
import jsonschema
import os
import json

def validate_schema(data: dict, root_schema_id: str, schema_folder: str) -> None:
    store: Dict[str, dict] = _load_schema_folder(schema_folder)
    for id, schema in store.items():
        jsonschema.Draft202012Validator.check_schema(schema)
    root_schema: dict = store[root_schema_id]
    resolver = jsonschema.RefResolver(base_uri="", referrer=root_schema, store=store)
    validator = jsonschema.Draft202012Validator(root_schema, resolver=resolver)
    validator.validate(data)

def _load_schema_folder(schema_folder: str) -> Dict[str, dict]:
    json_files: List[str] = []
    for dir_path, dir_names, file_names in os.walk(schema_folder):
        for file_name in file_names:
            if file_name.split(".")[-1].lower() == "json":
                json_files.append(os.path.join(dir_path, file_name))
    schema_store: Dict[str, str] = {}
    for json_file in json_files:
        with open(json_file, "r", encoding="utf-8") as f:
            loaded_json: dict = json.load(f)
        if "$schema" in loaded_json and "$id" in loaded_json:
            schema_store[loaded_json["$id"]] = loaded_json
    return schema_store



from dataclasses import dataclass
from typing import Dict, List, Tuple

from .PackageDataclasses import Tag

@dataclass
class TaxonomyDocument:
    priority: int
    prefix: str
    metadata: 'TaxonomyMetadata'
    namespaces: Dict[str, str]
    schema_imports: Dict[str, str]
    elements: List['TaxonomyElement']
    linkbase_imports: Dict[str, str]
    arc_roles_import: Dict[str, str]
    roles: List['TaxonomyRole']
    labels: Dict[str, List['LabelElement']]

    @property
    def rewrite_path(cls) -> List[str]:
        return [cls.metadata.publisher_url, "xbrl", cls.metadata.publication_date]

    @property
    def namespace(cls) -> str:
        return f"http://{'/'.join(cls.rewrite_path)}"
    
    @property
    def files_base_name(cls) -> str:
        return f"{cls.prefix}-{cls.metadata.publication_date}"

    @property
    def schema_url(cls) -> str:
        return f"{cls.namespace}/{cls.files_base_name}.xsd"
    
    @classmethod
    def from_dict(cls, data: dict) -> 'TaxonomyDocument':
        return cls(
            priority=data.get("priority", 0),
            prefix=data.get("prefix"), 
            metadata=TaxonomyMetadata.from_dict(data.get("metadata", {})),
            namespaces=data.get("namespaces", {}),
            schema_imports=data.get("schema_imports", {}),
            elements=[TaxonomyElement.from_dict(element_data) for element_data in data.get("elements", [])],
            linkbase_imports=data.get("linkbase_imports", {}),
            arc_roles_import=data.get("arc_roles_import", {}),
            roles=[TaxonomyRole.from_dict(role_data) for role_data in data.get("roles", [])],
            labels={labels_lang:[LabelElement.from_dict(label_element) for label_element in labels_data] for labels_lang, labels_data in data.get("labels", {}).items()}
        )

    def add_taxonomy(target_taxonomy, new_taxonomy: "TaxonomyDocument") -> Dict[str, str]:
        # add namespaces if not already known
        target_taxonomy.namespaces = { **new_taxonomy.namespaces, **target_taxonomy.namespaces}
        # add schemas to import
        target_taxonomy.schema_imports = { **new_taxonomy.schema_imports, **target_taxonomy.schema_imports}
        # add new elements, check for name and change it if it already exists with differnt attributes
        element_update_map: Dict[str, str] = {}
        for new_element in new_taxonomy.elements:
            used_elements: Dict[str, TaxonomyElement] = {element.name:element for element in target_taxonomy.elements}
            if not new_element.name in used_elements:
                target_taxonomy.elements.append(new_element)
            else:
                if not used_elements[new_element.name].equals(new_element):
                    index: int = 0
                    new_name: str = f"{new_element.name}_{index}"
                    while new_name in used_elements:
                        index += 1
                        new_name = f"{new_element.name}_{index}"
                    element_update_map[new_element.name] = new_name
                    new_element.name = new_name
                    target_taxonomy.elements.append(new_element)
        # import linkbases if not already imported
        target_taxonomy.linkbase_imports = { **new_taxonomy.linkbase_imports, **target_taxonomy.linkbase_imports}
        # import arc roles if not already imported
        target_taxonomy.arc_roles_import = { **new_taxonomy.arc_roles_import, **target_taxonomy.arc_roles_import}
        # update element references in roles by element_update_map
        for new_role in new_taxonomy.roles:
            new_role.update_elements(element_update_map)
        # merge roles
        target_roles: Dict[str, TaxonomyRole] = {role.href(""):role for role in target_taxonomy.roles}
        for new_role in new_taxonomy.roles:
            new_role_href: str = new_role.href("")
            if new_role_href in target_roles:
                # combine roles
                target_roles[new_role_href].merge(new_role)
            else:
                # add new role to taxonomy
                target_taxonomy.roles.append(new_role)
        # update element references in labels by element_update_map
        for label_lang in new_taxonomy.labels:
            for label_element in new_taxonomy.labels[label_lang]:
                label_element.update_element(element_update_map)
        # merge labels
        for label_lang in new_taxonomy.labels:
            if not label_lang in target_taxonomy.labels:
                target_taxonomy.labels[label_lang] = []
            for new_label_element in new_taxonomy.labels[label_lang]:
                target_label_elements: Dict[str, LabelElement] = {tar_label_ele.uname:tar_label_ele for tar_label_ele in target_taxonomy.labels[label_lang]}
                # if element not already used for labels, add the element
                if not new_label_element.uname in target_label_elements:
                    target_taxonomy.labels[label_lang].append(new_label_element)
                else:
                    # combine with the existing label
                    target_label_elements[new_label_element.uname].combine(new_label_element)
        return element_update_map

    def to_dict(cls) -> dict:
        return {
            "priority": cls.priority,
            "prefix": cls.prefix,
            "metadata": cls.metadata.to_dict(),
            "namespaces": cls.namespaces,
            "schema_imports": cls.schema_imports,
            "linkbase_imports": cls.linkbase_imports,
            "elements": [element.to_dict() for element in cls.elements],
            "arc_roles_import": cls.arc_roles_import,
            "roles": [role.to_dict() for role in cls.roles],
            "labels": {labels_lang:[label_element.to_dict() for label_element in labels_data] for labels_lang, labels_data in cls.labels.items()}
        }

@dataclass
class TaxonomyMetadata:
    name: str
    description: str
    publisher: str
    publisher_url: str
    publication_date: str
    publisher_country: str
    entrypoints: List['Entrypoint']

    @classmethod
    def from_dict(cls, data: dict) -> 'TaxonomyMetadata':
        return cls(
            name=data.get("name"),
            description=data.get("description"), 
            publisher=data.get("publisher"),
            publisher_url=data.get("publisher_url"),
            publisher_country=data.get("publisher_country"),
            publication_date=data.get("publication_date"),
            entrypoints=[Entrypoint.from_dict(entrypoint) for entrypoint in data.get("entrypoints", [])]
        )

    def to_dict(cls) -> dict:
        return {
            "name": cls.name,
            "description": cls.description,
            "publisher": cls.publisher,
            "publisher_url": cls.publisher_url,
            "publisher_country": cls.publisher_country,
            "publication_date": cls.publication_date,
            "entrypoints": [entrypoint.to_dict() for entrypoint in cls.entrypoints]
        }

@dataclass
class Entrypoint:
    name: str
    description: str
    documents: List[str]
    language: str

    @classmethod
    def from_dict(cls, data: dict) -> 'Entrypoint':
        return cls(
            name=data.get("name"),
            description=data.get("description"),
            documents=data.get("documents", []),
            language=data.get("language")
        )

    def to_dict(cls) -> dict:
        return {
            "name": cls.name,
            "description": cls.description,
            "documents": cls.documents,
            "language": cls.language
        }
    
@dataclass
class TaxonomyElement:
    balance: str
    period_type: str
    name: str
    nillable: bool
    abstract: bool
    substitution_group: Tag
    type: Tag

    @classmethod
    def from_dict(cls, data: dict) -> 'TaxonomyElement':
        return cls(
            balance=data.get("balance"),
            period_type=data.get("period_type"),
            name=data.get("name"),
            nillable=data.get("nillable"),
            abstract=data.get("abstract"),
            substitution_group=Tag.from_dict(data.get("substitution_group", {})),
            type=Tag.from_dict(data.get("type", {})),
        )

    def equals(cls, compare_element: "TaxonomyElement") -> bool:
        if cls.balance != compare_element.balance:
            return False
        if cls.period_type != compare_element.period_type:
            return False
        if cls.nillable != compare_element.nillable:
            return False
        if cls.abstract != compare_element.abstract:
            return False
        if cls.substitution_group.to_uname() != compare_element.substitution_group.to_uname():
            return False
        if cls.type.to_uname() != compare_element.type.to_uname():
            return False
        return True

    def to_dict(cls) -> dict:
        return {
            "balance": cls.balance,
            "period_type": cls.period_type,
            "name": cls.name,
            "nillable": cls.nillable,
            "abstract": cls.abstract,
            "substitution_group": cls.substitution_group.to_dict(),
            "type": cls.type.to_dict()
        }

@dataclass
class TaxonomyRole:
    role_name: str
    role_id: str
    role_uri: str
    schema_location: str
    presentation_linkbase: List['PresentationElement']
    definition_linkbase: List['DefinitionElement']
    calculation_linkbase: List['CalculationElement']

    @classmethod
    def from_dict(cls, data: dict) -> 'TaxonomyRole':
        return cls(
            role_name=data.get("role_name"),
            role_id=data.get("role_id"),
            role_uri=data.get("role_uri"),
            schema_location=data.get("schema_location"),
            presentation_linkbase=[PresentationElement.from_dict(element) for element in data.get("presentation_linkbase", [])],
            definition_linkbase=[DefinitionElement.from_dict(element) for element in data.get("definition_linkbase", [])],
            calculation_linkbase=[CalculationElement.from_dict(element) for element in data.get("calculation_linkbase", [])]
        )

    def uri(cls, taxonomy_namespace: str) -> str:
        if not cls.role_uri:
            return f"{taxonomy_namespace.rstrip('/')}/roles/{cls.role_id}"
        return cls.role_uri
    
    def href(cls, file_base_name: str) -> str:
        if not cls.schema_location:
            return f"{file_base_name}.xsd#{cls.role_id}"
        return f"{cls.schema_location}#{cls.role_id}"

    def update_elements(cls, update_element_map: Dict[str, str]) -> None:
        # update presentation
        for pres_element in cls.presentation_linkbase:
            pres_element.update_elements(update_element_map)
        # update definition
        for def_element in cls.definition_linkbase:
            def_element.update_elements(update_element_map)
        # update calculation
        for calc_element in cls.calculation_linkbase:
            calc_element.update_elements(update_element_map)

    def merge(target_role, new_role: "TaxonomyRole") -> None:
        # update presentation linkbase
        new_presentations: List[PresentationElement] = []
        for pres_item in new_role.presentation_linkbase:
            new_presentations += pres_item.deconstruct()
        extsting_pres_relations: List[Tuple[str, str]] = []
        for pres_item in target_role.presentation_linkbase:
            extsting_pres_relations += pres_item.get_relations()
        # clean presentations from the new ones which already exist
        for new_presentation in new_presentations:
            cleaned_children: List[PresentationElement] = []
            # add chold to cleaned children if the relation is not alredy known
            for child in new_presentation.children:
                if not child.get_relation(new_presentation) in extsting_pres_relations:
                    cleaned_children.append(child)
            # replace role children with cleaned and if there are some left add it to the presentation linkbase
            new_presentation.children = cleaned_children
            if new_presentation.children:
                target_role.presentation_linkbase.append(new_presentation)
        # update defintion linkbase
        new_definitions: List[DefinitionElement] = []
        for def_item in new_role.definition_linkbase:
            new_definitions += def_item.deconstruct()
        existing_def_relations: List[Tuple[str, str]] = []
        for def_item in target_role.definition_linkbase:
            existing_def_relations += def_item.get_relations()
        # clean definition from the new ones which already exist
        for new_definition in new_definitions:
            cleaned_children: List[DefinitionElement] = []
            # add child to cleaned children if the relation is not alredy known
            for child in new_definition.children:
                if not child.get_relation(new_definition) in existing_def_relations:
                    cleaned_children.append(child)
            # replace role children and if there are children add to presentation
            new_definition.children = cleaned_children
            if new_definition.children:
                target_role.definition_linkbase.append(new_definition)
        # update calculation linkbase
        new_calculations: List[CalculationElement] = []
        for calc_item in new_role.calculation_linkbase:
            new_calculations += calc_item.deconstruct()
        extsting_calc_relations: List[Tuple[str, str]] = []
        for calc_item in target_role.calculation_linkbase:
            extsting_calc_relations += calc_item.get_relations()
        # if a calculation for a element already exists, do not interfere with it
        parent_calculation_ids: List[str] = [relation[0] for relation in extsting_calc_relations]
        for new_calc in new_calculations:
            if not new_calc.id in parent_calculation_ids:
                target_role.calculation_linkbase.append(new_calc)

    def to_dict(cls) -> dict:
        return {
            "role_name": cls.role_name,
            "role_id": cls.role_id,
            "schema_location": cls.schema_location,
            "presentation_linkbase": [element.to_dict() for element in cls.presentation_linkbase],
            "definition_linkbase": [element.to_dict() for element in cls.definition_linkbase],
            "calculation_linkbase": [element.to_dict() for element in cls.calculation_linkbase]
        }

@dataclass
class LinkbaseElement:
    element_id: str
    schema_location: str
    arc_role: str
    children: List['LinkbaseElement']

    @classmethod
    def from_dict(cls, data: dict) -> 'LinkbaseElement':
        return cls(
            element_id=data.get("element_id"),
            schema_location=data.get("schema_location"),
            arc_role=data.get("arc_role"),
            children=[LinkbaseElement.from_dict(child) for child in data.get("children", [])]
        )
    
    @property
    def id(cls) -> str:
        return f"{{{cls.schema_location}}}{cls.element_id}"

    @property
    def extended_id(cls) -> str:
        return cls.id
    
    def get_relation(cls, parent: "LinkbaseElement") -> Tuple[str, str]:
        return (parent.id, cls.extended_id)

    def get_relations(cls) -> List[Tuple[str, str]]:
        relationships: List[Tuple[str, str]] = []
        if cls.children:
            for child in cls.children:
                relationships.append(child.get_relation(cls))
                relationships += child.get_relations()
        return relationships

    def copy(cls, include_children: bool = True) -> "LinkbaseElement":
        raise Exception("The copy function was not implemented on the Base LinkbaseElement Class!")
    
    def deconstruct(cls) -> List["LinkbaseElement"]:
        raise Exception("The deconstruct function was not implemented on the Base LinkbaseElement Class!")
    
    def update_elements(cls, update_element_map: Dict[str, str]) -> None:
        if not cls.schema_location and cls.element_id in update_element_map:
            cls.element_id = update_element_map[cls.element_id]
        for child in cls.children:
            child.update_elements(update_element_map)
    
    def to_dict(cls) -> dict:
        return {
            "element_id": cls.element_id,
            "schema_location": cls.schema_location,
            "arc_role": cls.arc_role,
            "children": [child.to_dict() for child in cls.children]
        }
    
@dataclass
class PresentationElement(LinkbaseElement):
    order: int
    preferred_label: str
    children: List['PresentationElement']

    @classmethod
    def from_dict(cls, data: dict) -> 'PresentationElement':
        return cls(
            element_id=data.get("element_id"),
            schema_location=data.get("schema_location"),
            arc_role=data.get("arc_role"),
            order=data.get("order", 0),
            preferred_label=data.get("preferred_label"),
            children=[PresentationElement.from_dict(child) for child in data.get("children", [])]
        )

    @property
    def extended_id(cls) -> str:
        return f"{cls.id}_{cls.preferred_label}"

    def copy(cls, include_children: bool = True, include_sub_children: bool = True, include_all_children: bool = True) -> "PresentationElement":
        children: List[PresentationElement] = []
        if include_children:
            children = [child.copy(include_sub_children, include_all_children, include_all_children) for child in cls.children]
        return PresentationElement(
            element_id=cls.element_id,
            schema_location=cls.schema_location,
            arc_role=cls.arc_role,
            order=cls.order,
            preferred_label=cls.preferred_label,
            children=children
        )

    def deconstruct(cls) -> List["PresentationElement"]:
        deconstructed: List["PresentationElement"] = []
        if cls.children:
            deconstructed = [cls.copy(True, False, False)]
            for child in cls.children:
                deconstructed += child.deconstruct()
        return deconstructed

    def to_dict(cls) -> dict:
        return {
            "element_id": cls.element_id,
            "schema_location": cls.schema_location,
            "arc_role": cls.arc_role,
            "order": cls.order,
            "preferred_label": cls.preferred_label,
            "children": [child.to_dict() for child in cls.children]
        }

@dataclass
class CalculationElement(LinkbaseElement):
    weight: int
    children: List['CalculationElement']

    @classmethod
    def from_dict(cls, data: dict) -> 'CalculationElement':
        return cls(
            element_id=data.get("element_id"),
            schema_location=data.get("schema_location"),
            arc_role=data.get("arc_role"),
            weight=data.get("weight", 0),
            children=[CalculationElement.from_dict(child) for child in data.get("children", [])]
        )

    def copy(cls, include_children: bool = True, include_sub_children: bool = True, include_all_children: bool = True) -> "CalculationElement":
        children: List[CalculationElement] = []
        if include_children:
            children = [child.copy(include_sub_children, include_all_children, include_all_children) for child in cls.children]
        return CalculationElement(
            element_id=cls.element_id,
            schema_location=cls.schema_location,
            arc_role=cls.arc_role,
            weight=cls.weight,
            children=children
        )

    def deconstruct(cls) -> List["CalculationElement"]:
        deconstructed: List["CalculationElement"] = []
        if cls.children:
            deconstructed = [cls.copy(True, False, False)]
            for child in cls.children:
                deconstructed += child.deconstruct()
        return deconstructed

    def to_dict(cls) -> dict:
        return {
            "element_id": cls.element_id,
            "schema_location": cls.schema_location,
            "arc_role": cls.arc_role,
            "weight": cls.weight,
            "children": [child.to_dict() for child in cls.children]
        }
    
@dataclass
class DefinitionElement(LinkbaseElement):
    context_element: str
    closed: bool
    children: List['DefinitionElement']

    @classmethod
    def from_dict(cls, data: dict) -> 'DefinitionElement':
        return cls(
            element_id=data.get("element_id"),
            schema_location=data.get("schema_location"),
            arc_role=data.get("arc_role"),
            context_element=data.get("context_element"),
            closed=data.get("closed"),
            children=[DefinitionElement.from_dict(child) for child in data.get("children", [])]
        )
    
    @property
    def extended_id(cls) -> str:
        return f"{cls.id}_{cls.arc_role}"

    def copy(cls, include_children: bool = True, include_sub_children: bool = True, include_all_children: bool = True) -> "DefinitionElement":
        children: List[DefinitionElement] = []
        if include_children:
            children = [child.copy(include_sub_children, include_all_children, include_all_children) for child in cls.children]
        return DefinitionElement(
            element_id=cls.element_id,
            schema_location=cls.schema_location,
            arc_role=cls.arc_role,
            context_element=cls.context_element,
            closed=cls.closed,
            children=children
        )

    def deconstruct(cls) -> List["DefinitionElement"]:
        deconstructed: List["DefinitionElement"] = []
        if cls.children:
            deconstructed = [cls.copy(True, False, False)]
            for child in cls.children:
                deconstructed += child.deconstruct()
        return deconstructed

    def to_dict(cls) -> dict:
        return {
            "element_id": cls.element_id,
            "schema_location": cls.schema_location,
            "arc_role": cls.arc_role,
            "context_element": cls.context_element,
            "closed": cls.closed,
            "children": [child.to_dict() for child in cls.children]
        }

@dataclass
class LabelElement:
    element_id: str
    schema_location: str
    labels: List['LabelData']
    
    @classmethod
    def from_dict(cls, data: dict) -> 'LabelElement':
        return cls(
            element_id=data.get("element_id"),
            schema_location=data.get("schema_location"),
            labels=[LabelData.from_dict(label_data) for label_data in data.get("labels", [])]
        )
    
    def update_element(cls, update_element_map: Dict[str, str]) -> None:
        if not cls.schema_location and cls.element_id in update_element_map:
            cls.element_id = update_element_map[cls.element_id]
    
    def combine(cls, new_label_element: "LabelElement") -> None:
        used_label_roles: List[str] = [label_data.label_role for label_data in cls.labels]
        for new_label in new_label_element.labels:
            if not new_label.label_role in used_label_roles:
                cls.labels.append(new_label)

    @property
    def uname(cls) -> str:
        return f"{{{cls.schema_location}}}{cls.element_id}"

    def to_dict(cls) -> dict:
        return {
            "element_id": cls.element_id,
            "schema_location": cls.schema_location,
            "labels": [label_data.to_dict() for label_data in cls.labels]
        }

@dataclass
class LabelData:
    label_role: str
    label: str

    @classmethod
    def from_dict(cls, data: dict) -> 'LabelData':
        return cls(
            label_role=data.get("label_role"),
            label=data.get("label")
        )
    
    def to_dict(cls) -> dict:
        return {
            "label_role": cls.label_role,
            "label": cls.label
        }


from typing import Dict, List, Tuple
from lxml import etree
import json

from .utils import xml_to_string
from .TaxonomyDataclasses import CalculationElement, DefinitionElement, PresentationElement, TaxonomyDocument, TaxonomyElement
from .PackageDataclasses import File, Tag

TAXONOMY_PACKAGE_NAMESPACE: str = "http://xbrl.org/2016/taxonomy-package"
TAXONOMY_PACKAGE_PREFIX: str = "tp"
XML_NAMESPACE: str = "http://www.w3.org/XML/1998/namespace"
XML_PREFIX: str = "xml"
XSI_NAMESPACE: str = "http://www.w3.org/2001/XMLSchema-instance"
XSI_PREFIX: str = "xsi"
LINKBASE_NAMESPACE: str = "http://www.xbrl.org/2003/linkbase"
LINKBASE_PREFIX: str = "link"
XLINK_NAMESPACE: str = "http://www.w3.org/1999/xlink"
XLINK_PREFIX: str = "xlink"
XBRLI_NAMESPACE: str = "http://www.xbrl.org/2003/instance"
XBRLI_PREFIX: str = "xbrli"
SCHEMA_NAMESPACE: str = "http://www.w3.org/2001/XMLSchema"
SCHEMA_PREFIX: str = "xs"

class TaxonomyProducer:
    taxonomy_document: TaxonomyDocument

    def __init__(cls, document: TaxonomyDocument):
        cls.taxonomy_document = document

    def create_files(cls, reports_folder: File = None, package_extension: str = "zip") -> File:
        # create base folder structure
        root_folder = File(name=f'{"_".join(cls.taxonomy_document.metadata.name.split())}', zip_extension=package_extension)
        if reports_folder:
            root_folder.contained_files.append(reports_folder)
        # create taxonomy files
        taxonomy_folder = root_folder
        for part in cls.taxonomy_document.rewrite_path:
            parent_folder = taxonomy_folder
            taxonomy_folder = File(name=part)
            parent_folder.contained_files.append(taxonomy_folder)
        cls._create_taxonomy_files(taxonomy_folder)
        # create meta information
        meta_inf_folder = File(name="META-INF", contained_files=cls._create_meta_inf_files(package_extension if reports_folder else None))
        root_folder.contained_files.append(meta_inf_folder)
        return root_folder

    def _create_meta_inf_files(cls, package_extension: str = None) -> List[File]:
        # create catalog file
        catalog_namespace: str = "urn:oasis:names:tc:entity:xmlns:xml:catalog"
        catalog_namespace_map = {
            None: catalog_namespace
        }
        catalog_root: etree.Element = etree.Element(f"{{{catalog_namespace}}}catalog", nsmap=catalog_namespace_map)
        rewrite_element: etree.Element = etree.SubElement(
            catalog_root,
            f"{{{catalog_namespace}}}rewriteURI",
            {
                "rewritePrefix": f"../{'/'.join(cls.taxonomy_document.rewrite_path)}/",
                "uriStartString": f"{cls.taxonomy_document.namespace}/"
            }
        ) 

        # create taxonomyPackage file
        tp_namespace_map = {
            TAXONOMY_PACKAGE_PREFIX: TAXONOMY_PACKAGE_NAMESPACE,
            XSI_PREFIX: XSI_NAMESPACE,
            XML_PREFIX: XML_NAMESPACE
        }
        taxonomy_package_element: etree.Element = etree.Element(
            f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}taxonomyPackage",
            {
                f"{{{XML_NAMESPACE}}}lang": "en",
                f"{{{XSI_NAMESPACE}}}schemaLocation": "http://xbrl.org/2016/taxonomy-package http://www.xbrl.org/2016/taxonomy-package.xsd"
            },
            nsmap=tp_namespace_map
        ) 
        identifier_element: etree.Element = etree.SubElement(
            taxonomy_package_element,
            f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}identifier"
        )
        identifier_element.text = cls.taxonomy_document.namespace
        name_element: etree.Element = etree.SubElement(
            taxonomy_package_element,
            f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}name"
        )
        name_element.text = cls.taxonomy_document.metadata.name
        description_element: etree.Element = etree.SubElement(
            taxonomy_package_element,
            f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}description"
        )
        description_element.text = cls.taxonomy_document.metadata.description
        version_element: etree.Element = etree.SubElement(
            taxonomy_package_element,
            f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}version"
        )
        version_element.text = cls.taxonomy_document.metadata.publication_date
        publisher_element: etree.Element = etree.SubElement(
            taxonomy_package_element,
            f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}publisher"
        )
        publisher_element.text = cls.taxonomy_document.metadata.publisher
        publisherUrl_element: etree.Element = etree.SubElement(
            taxonomy_package_element,
            f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}publisherURL"
        )
        publisherUrl_element.text = cls.taxonomy_document.metadata.publisher_url
        publisherCountry_element: etree.Element = etree.SubElement(
            taxonomy_package_element,
            f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}publisherCountry"
        )
        publisherCountry_element.text = cls.taxonomy_document.metadata.publisher_country
        publicationdate_element: etree.Element = etree.SubElement(
            taxonomy_package_element,
            f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}publicationDate"
        )
        publicationdate_element.text = cls.taxonomy_document.metadata.publication_date
        entrypoints_element: etree.Element = etree.SubElement(
            taxonomy_package_element,
            f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}entryPoints"
        )
        for entrypoint in cls.taxonomy_document.metadata.entrypoints:
            entrypoint_element: etree.Element = etree.SubElement(
                entrypoints_element,
                f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}entryPoint"
            )
            ep_name_element: etree.Element = etree.SubElement(
                entrypoint_element,
                f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}name"
            )
            ep_name_element.text = entrypoint.name
            ep_description_element: etree.Element = etree.SubElement(
                entrypoint_element,
                f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}description"
            )
            ep_description_element.text = entrypoint.description
            for entrypoint_document in [cls.taxonomy_document.schema_url] + entrypoint.documents:
                ep_document_element: etree.Element = etree.SubElement(
                    entrypoint_element,
                    f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}entryPointDocument",
                    {
                        "href": entrypoint_document
                    }
                )
            ep_langs_element: etree.Element = etree.SubElement(
                entrypoint_element,
                f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}languages"
            )
            ep_lang_element: etree.Element = etree.SubElement(
                ep_langs_element,
                f"{{{TAXONOMY_PACKAGE_NAMESPACE}}}language"
            )
            ep_lang_element.text = entrypoint.language

        return_files: List[File] = [
            File(
                name="catalog.xml", 
                content=xml_to_string(
                    catalog_root, 
                    doctype='<!DOCTYPE catalog PUBLIC "-//OASIS//DTD Entity Resolution XML Catalog V1.0//EN" "http://www.oasis-open.org/committees/entity/release/1.0/catalog.dtd">'
                )
            ),
            File(
                name="taxonomyPackage.xml", 
                content=xml_to_string(
                    taxonomy_package_element
                )
            )
        ]

        if package_extension:
            supported_package_extensions: Dict[str, str] = {
                "zip": "https://xbrl.org/report-package/2023",
                "xbri": "https://xbrl.org/report-package/2023/xbri",
                "xbr": "https://xbrl.org/report-package/2023/xbr"
            }
            return_files.append(File(
                name="reportPackage.json",
                content=json.dumps({
                    "documentInfo": {
                        "documentType": supported_package_extensions.get(package_extension)
                    }
                })
            ))
        return return_files

    def _create_taxonomy_files(cls, taxonomy_folder: File) -> None:
        cls._create_schema(taxonomy_folder)
        cls._create_linkbase(
            f"{cls.taxonomy_document.files_base_name}_pre.xml",
            taxonomy_folder,
            cls._create_presentation()
        )
        cls._create_linkbase(
            f"{cls.taxonomy_document.files_base_name}_cal.xml",
            taxonomy_folder,
            cls._create_calculation()
        )
        cls._create_linkbase(
            f"{cls.taxonomy_document.files_base_name}_def.xml",
            taxonomy_folder,
            cls._create_definition()
        )
        for labels_lang in cls.taxonomy_document.labels:
            cls._create_linkbase(
                f"{cls.taxonomy_document.files_base_name}_lab-{labels_lang}.xml",
                taxonomy_folder,
                cls._create_label_linkbase(labels_lang)   
            )

    def _create_schema(cls, taxonomy_folder: File) -> None:
        # create base structure
        namespace_map = {
            cls.taxonomy_document.prefix: cls.taxonomy_document.namespace,
            XLINK_PREFIX: XLINK_NAMESPACE,
            LINKBASE_PREFIX: LINKBASE_NAMESPACE,
            XBRLI_PREFIX: XBRLI_NAMESPACE,
            SCHEMA_PREFIX: SCHEMA_NAMESPACE
        }
        for namespace, prefix in cls.taxonomy_document.namespaces.items():
            namespace_map[prefix] = namespace
        schema_root: etree.Element = etree.Element(
            f"{{{SCHEMA_NAMESPACE}}}schema",
            {
                "targetNamespace": cls.taxonomy_document.namespace
            },
            nsmap=namespace_map
        )
        # import schemas
        for import_schema_ns, import_schema_location in cls.taxonomy_document.schema_imports.items():
            schema_import_element: etree.Element = etree.SubElement(
                schema_root,
                f"{{{SCHEMA_NAMESPACE}}}import",
                {
                    "schemaLocation": import_schema_location,
                    "namespace": import_schema_ns
                }
            )
        # creat annotation elements
        annotation_element: etree.Element = etree.SubElement(
            schema_root,
            f"{{{SCHEMA_NAMESPACE}}}annotation"
        )
        appinfo_element: etree.Element = etree.SubElement(
            annotation_element,
            f"{{{SCHEMA_NAMESPACE}}}appinfo"
        )
        # import taxonomy linkbases
        linkbases: Dict[str, str] = {
            f"{cls.taxonomy_document.files_base_name}_pre.xml": "http://www.xbrl.org/2003/role/presentationLinkbaseRef",
            f"{cls.taxonomy_document.files_base_name}_def.xml": "http://www.xbrl.org/2003/role/definitionLinkbaseRef",
            f"{cls.taxonomy_document.files_base_name}_cal.xml": "http://www.xbrl.org/2003/role/calculationLinkbaseRef"
        }
        for lang in cls.taxonomy_document.labels:
            linkbases[f"{cls.taxonomy_document.files_base_name}_lab-{lang}.xml"] = "http://www.xbrl.org/2003/role/labelLinkbaseRef"
        linkbases.update(cls.taxonomy_document.linkbase_imports)
        for linkbase_href, linkbase_role in linkbases.items():
            linkbase_ref_attributes = {
                f"{{{XLINK_NAMESPACE}}}arcrole": "http://www.w3.org/1999/xlink/properties/linkbase",
                f"{{{XLINK_NAMESPACE}}}href": linkbase_href,
                f"{{{XLINK_NAMESPACE}}}type": "simple"
            }
            if linkbase_role:
                linkbase_ref_attributes[f"{{{XLINK_NAMESPACE}}}role"] = linkbase_role
            linkbase_element: etree.Element = etree.SubElement(
                appinfo_element,
                f"{{{LINKBASE_NAMESPACE}}}linkbaseRef",
                linkbase_ref_attributes
            )
        # add roles
        for role in cls.taxonomy_document.roles:
            if not role.schema_location:
                role_element: etree.Element = etree.SubElement(
                    appinfo_element,
                    f"{{{LINKBASE_NAMESPACE}}}roleType",
                    {
                        "id": role.role_id,
                        "roleURI": role.uri(cls.taxonomy_document.namespace)
                    }
                )
                role_name_element: etree.Element = etree.SubElement(
                    role_element,
                    f"{{{LINKBASE_NAMESPACE}}}definition"
                )
                role_name_element.text = role.role_name
                if role.presentation_linkbase:
                    role_link_element: etree.Element = etree.SubElement(
                        role_element,
                        f"{{{LINKBASE_NAMESPACE}}}usedOn"
                    )
                    role_link_element.text = "link:presentationLink"
                if role.calculation_linkbase:
                    role_link_element: etree.Element = etree.SubElement(
                        role_element,
                        f"{{{LINKBASE_NAMESPACE}}}usedOn"
                    )
                    role_link_element.text = "link:calculationLink"
                if role.definition_linkbase:
                    role_link_element: etree.Element = etree.SubElement(
                        role_element,
                        f"{{{LINKBASE_NAMESPACE}}}usedOn"
                    )
                    role_link_element.text = "link:definitionLink"
                role_link_element: etree.Element = etree.SubElement(
                    role_element,
                    f"{{{LINKBASE_NAMESPACE}}}usedOn"
                )
                role_link_element.text = "link:labelLink"
        # add taxonomy elements
        for element_data in cls.taxonomy_document.elements:
            attributes: Dict[str, str] = {
                    "id": element_data.name,
                    f"{{{XBRLI_NAMESPACE}}}periodType": element_data.period_type,
                    "name": element_data.name,
                    "nillable": "true" if element_data.nillable else "false",
                    "substitutionGroup": element_data.substitution_group.to_prefixed_name(cls.taxonomy_document.namespaces),
                    "type": element_data.type.to_prefixed_name(cls.taxonomy_document.namespaces)
            }
            if element_data.abstract:
                attributes["abstract"] = "true"
            if element_data.balance:
                attributes[f"{{{XBRLI_NAMESPACE}}}balance"] = element_data.balance
            element_element: etree.Element = etree.SubElement(
                schema_root,
                f"{{{SCHEMA_NAMESPACE}}}element",
                attributes
            ) 
        taxonomy_folder.contained_files.append(File(f"{cls.taxonomy_document.files_base_name}.xsd", content=xml_to_string(schema_root)))

    def _create_linkbase(cls, name: str, taxonomy_folder: File, content_data: Tuple[List[etree.Element], List[str]]) -> None:
        content_elements, used_arcroles = content_data
        namespace_map = {
            LINKBASE_PREFIX: LINKBASE_NAMESPACE,
            XSI_PREFIX: XSI_NAMESPACE,
            XLINK_PREFIX: XLINK_NAMESPACE,
            XML_PREFIX: XML_NAMESPACE
        }
        linkbase_root: etree.Element = etree.Element(
            f"{{{LINKBASE_NAMESPACE}}}linkbase",
            {
                f"{{{XSI_NAMESPACE}}}schemaLocation": "http://www.xbrl.org/2003/linkbase http://www.xbrl.org/2003/xbrl-linkbase-2003-12-31.xsd"
            },
            nsmap=namespace_map
        )
        # import arc roles if needed
        for arc_role_uri, arc_role_href in cls.taxonomy_document.arc_roles_import.items():
            if arc_role_uri in used_arcroles:
                arcrole_ref_element: etree.Element = etree.SubElement(
                    linkbase_root,
                    f"{{{LINKBASE_NAMESPACE}}}arcroleRef",
                    {
                        "arcroleURI": arc_role_uri,
                        f"{{{XLINK_NAMESPACE}}}href": arc_role_href,
                        f"{{{XLINK_NAMESPACE}}}type": "simple"
                    }
                )
        # add content of linkbase
        for content_element in content_elements:
            linkbase_root.append(content_element)
        taxonomy_folder.contained_files.append(File(name, content=xml_to_string(linkbase_root)))

    def _create_presentation(cls) -> Tuple[List[etree.Element], List[str]]:
        # create presentation Linkbase content
        content_elements: List[etree.Element] = []
        used_arc_roles: List[str] = []
        for role_data in cls.taxonomy_document.roles:
            if role_data.presentation_linkbase:
                role_ref_element: etree.Element = etree.Element(
                    f"{{{LINKBASE_NAMESPACE}}}roleRef",
                    {
                        "roleURI": role_data.uri(cls.taxonomy_document.namespace),
                        f"{{{XLINK_NAMESPACE}}}href": role_data.href(cls.taxonomy_document.files_base_name),
                        f"{{{XLINK_NAMESPACE}}}type": "simple"
                    }
                ) 
                content_elements.append(role_ref_element)
                presentation_link_element: etree.Element = etree.Element(
                    f"{{{LINKBASE_NAMESPACE}}}presentationLink",
                    {
                        f"{{{XLINK_NAMESPACE}}}role": role_data.uri(cls.taxonomy_document.namespace),
                        f"{{{XLINK_NAMESPACE}}}type": "extended"
                    }
                )
                # add elements to the presentation linkbase
                locators: Dict[str, str] = {}
                for child in role_data.presentation_linkbase:
                    locators, child_arc_roles = cls._add_presentation_item(
                        child, 
                        presentation_link_element, 
                        locators
                    )
                    used_arc_roles += child_arc_roles
                content_elements.append(presentation_link_element)
        return content_elements, used_arc_roles
        
    def _add_presentation_item(cls, presentation_element: PresentationElement, parent_element: etree.Element, locators: Dict[str, str], parent_element_locator: str = None) -> Tuple[Dict[str, str], List[str]]:
        used_arcroles: List[str] = []
        element_label, locators = cls._add_element_locator(
            parent_element,
            presentation_element.schema_location,
            presentation_element.element_id,
            locators
        )
        if parent_element_locator:
            arc_attributes = {
                f"{{{XLINK_NAMESPACE}}}type": "arc",
                f"{{{XLINK_NAMESPACE}}}from": parent_element_locator,
                f"{{{XLINK_NAMESPACE}}}to": element_label,
                f"{{{XLINK_NAMESPACE}}}arcrole": presentation_element.arc_role,
                "order": str(presentation_element.order)
            }
            if presentation_element.arc_role:
                used_arcroles.append(presentation_element.arc_role)
            if presentation_element.preferred_label:
                arc_attributes["preferredLabel"] = presentation_element.preferred_label
            presentation_arc_element: etree.Element = etree.SubElement(
                parent_element,
                f"{{{LINKBASE_NAMESPACE}}}presentationArc",
                arc_attributes
            )
        for child in presentation_element.children:
            locators, child_arc_roles = cls._add_presentation_item(
                child, 
                parent_element, 
                locators,
                element_label
            )
            used_arcroles += child_arc_roles
        return locators, used_arcroles
    
    def _create_calculation(cls) -> Tuple[List[etree.Element], List[str]]:
        # create calculation Linkbase
        content_elements: List[etree.Element] = []
        used_arc_roles: List[str] = []
        for role_data in cls.taxonomy_document.roles:
            if role_data.calculation_linkbase:
                role_ref_element: etree.Element = etree.Element(
                    f"{{{LINKBASE_NAMESPACE}}}roleRef",
                    {
                        "roleURI": role_data.uri(cls.taxonomy_document.namespace),
                        f"{{{XLINK_NAMESPACE}}}href": role_data.href(cls.taxonomy_document.files_base_name),
                        f"{{{XLINK_NAMESPACE}}}type": "simple"
                    }
                ) 
                content_elements.append(role_ref_element)
                calculation_link_element: etree.Element = etree.Element(
                    f"{{{LINKBASE_NAMESPACE}}}calculationLink",
                    {
                        f"{{{XLINK_NAMESPACE}}}role": role_data.uri(cls.taxonomy_document.namespace),
                        f"{{{XLINK_NAMESPACE}}}type": "extended"
                    }
                )
                # add elements to the calculation linkbase
                locators: Dict[str, str] = {}
                for child in role_data.calculation_linkbase:
                    locators, child_arc_roles = cls._add_calculation_item(
                        child, 
                        calculation_link_element, 
                        locators
                    )
                    used_arc_roles += child_arc_roles
                content_elements.append(calculation_link_element)
        return content_elements, used_arc_roles

    def _add_calculation_item(cls, calculation_element: CalculationElement, parent_element: etree.Element, locators: Dict[str, str], parent_element_locator: str = None) -> Tuple[Dict[str, str], List[str]]:
        used_arcroles: List[str] = []
        element_label, locators = cls._add_element_locator(
            parent_element,
            calculation_element.schema_location,
            calculation_element.element_id,
            locators
        )
        if parent_element_locator:
            arc_attributes = {
                f"{{{XLINK_NAMESPACE}}}type": "arc",
                f"{{{XLINK_NAMESPACE}}}from": parent_element_locator,
                f"{{{XLINK_NAMESPACE}}}to": element_label,
                f"{{{XLINK_NAMESPACE}}}arcrole": calculation_element.arc_role,
                "weight": str(calculation_element.weight)
            }
            if calculation_element.arc_role:
                used_arcroles.append(calculation_element.arc_role)
            calculation_arc_element: etree.Element = etree.SubElement(
                parent_element,
                f"{{{LINKBASE_NAMESPACE}}}calculationArc",
                arc_attributes
            )
        for child in calculation_element.children:
            locators, child_arc_roles = cls._add_calculation_item(
                child, 
                parent_element, 
                locators,
                element_label
            )
            used_arcroles += child_arc_roles
        return locators, used_arcroles

    def _create_definition(cls) -> Tuple[List[etree.Element], List[str]]:
        # create definition Linkbase
        content_elements: List[etree.Element] = []
        used_arc_roles: List[str] = []
        for role_data in cls.taxonomy_document.roles:
            if role_data.definition_linkbase:
                role_ref_element: etree.Element = etree.Element(
                    f"{{{LINKBASE_NAMESPACE}}}roleRef",
                    {
                        "roleURI": role_data.uri(cls.taxonomy_document.namespace),
                        f"{{{XLINK_NAMESPACE}}}href": role_data.href(cls.taxonomy_document.files_base_name),
                        f"{{{XLINK_NAMESPACE}}}type": "simple"
                    }
                )
                content_elements.append(role_ref_element)
                definition_link_element: etree.Element = etree.Element(
                    f"{{{LINKBASE_NAMESPACE}}}definitionLink",
                    {
                        f"{{{XLINK_NAMESPACE}}}role": role_data.uri(cls.taxonomy_document.namespace),
                        f"{{{XLINK_NAMESPACE}}}type": "extended"
                    }
                )
                # add elements to the calculation linkbase
                locators: Dict[str, str] = {}
                for child in role_data.definition_linkbase:
                    locators, child_arc_roles = cls._add_definition_item(
                        child, 
                        definition_link_element, 
                        locators
                    )
                    used_arc_roles += child_arc_roles
                content_elements.append(definition_link_element)
        return content_elements, used_arc_roles
    
    def _add_definition_item(cls, definition_element: DefinitionElement, parent_element: etree.Element, locators: Dict[str, str], parent_element_locator: str = None) -> Tuple[Dict[str, str], List[str]]:
        used_arcroles: List[str] = []
        element_label, locators = cls._add_element_locator(
            parent_element,
            definition_element.schema_location,
            definition_element.element_id,
            locators
        )
        if parent_element_locator:
            xbrldt_namespace = "http://xbrl.org/2005/xbrldt"
            nsmp = {
                "xbrldt": xbrldt_namespace
            }
            arc_attributes = {
                f"{{{XLINK_NAMESPACE}}}type": "arc",
                f"{{{XLINK_NAMESPACE}}}from": parent_element_locator,
                f"{{{XLINK_NAMESPACE}}}to": element_label,
                f"{{{XLINK_NAMESPACE}}}arcrole": definition_element.arc_role
            }
            if definition_element.arc_role:
                used_arcroles.append(definition_element.arc_role)
            if definition_element.closed != None:
                arc_attributes[f"{{{xbrldt_namespace}}}closed"] = "true" if definition_element.closed else "false"
            if definition_element.context_element != None:
                arc_attributes[f"{{{xbrldt_namespace}}}contextElement"] = definition_element.context_element
            definition_arc_element: etree.Element = etree.SubElement(
                parent_element,
                f"{{{LINKBASE_NAMESPACE}}}definitionArc",
                arc_attributes,
                nsmap=nsmp
            )
        for child in definition_element.children:
            locators, child_arc_roles = cls._add_definition_item(
                child, 
                parent_element, 
                locators,
                element_label
            )
            used_arcroles += child_arc_roles
        return locators, used_arcroles

    def _create_label_linkbase(cls, labels_lang: str) -> Tuple[List[etree.Element], List[str]]:
        label_elements = cls.taxonomy_document.labels[labels_lang]
        label_link_element: etree.Element = etree.Element(
            f"{{{LINKBASE_NAMESPACE}}}labelLink",
            {
                f"{{{XLINK_NAMESPACE}}}role": "http://www.xbrl.org/2003/role/link",
                f"{{{XLINK_NAMESPACE}}}type": "extended"
            }
        )
        locators: Dict[str, str] = {}
        label_id: int = 0
        for label_element_data in label_elements:
            locator_label, locators = cls._add_element_locator(
                label_link_element,
                label_element_data.schema_location,
                label_element_data.element_id,
                locators
            )
            for label_data in label_element_data.labels:
                label_label: str = f"label_{label_id}"
                label_id += 1
                label_element: etree.Element = etree.SubElement(
                    label_link_element,
                    f"{{{LINKBASE_NAMESPACE}}}label",
                    {
                        f"{{{XLINK_NAMESPACE}}}label": label_label,
                        f"{{{XLINK_NAMESPACE}}}role": label_data.label_role,
                        f"{{{XLINK_NAMESPACE}}}type": "resource",
                        f"{{{XML_NAMESPACE}}}lang": labels_lang
                    }
                )
                label_element.text = label_data.label
                label_arc_element: etree.Element = etree.SubElement(
                    label_link_element,
                    f"{{{LINKBASE_NAMESPACE}}}labelArc",
                    {
                        f"{{{XLINK_NAMESPACE}}}type": "arc",
                        f"{{{XLINK_NAMESPACE}}}from": locator_label,
                        f"{{{XLINK_NAMESPACE}}}to": label_label,
                        f"{{{XLINK_NAMESPACE}}}arcrole": "http://www.xbrl.org/2003/arcrole/concept-label",
                        "priority": "100"
                    }
                )
        return [label_link_element], []
        
    def _add_element_locator(cls, parent_element: etree.Element, schema_location: str, element_id: str, located_elements_elements: Dict[str, str]) -> None:
        if schema_location:
            locator_href = f"{schema_location}#{element_id}"
        else:
            locator_href = f"{cls.taxonomy_document.files_base_name}.xsd#{element_id}"
        if locator_href in located_elements_elements:
            return located_elements_elements[locator_href], located_elements_elements
        locator_label: str = f"locator_{len(located_elements_elements.keys())}"
        element_locator: etree.Element = etree.SubElement(
            parent_element,
            f"{{{LINKBASE_NAMESPACE}}}loc",
            {
                f"{{{XLINK_NAMESPACE}}}type": "locator",
                f"{{{XLINK_NAMESPACE}}}href": locator_href,
                f"{{{XLINK_NAMESPACE}}}label": locator_label
            }
        )
        located_elements_elements[locator_href] = locator_label
        return locator_label, located_elements_elements


import os
from typing import Dict, List

from ..xbrl_generation.ContentDataclasses import ContentDocument, ContentItem
from ..xbrl_generation.InputData import InputData
from .BaseLoader import BaseLoader
from .DocxLoader import DocxLoader

DOCUMENT_LOADERS: Dict[str, BaseLoader] = {
    "docx": DocxLoader()
}

def doc_to_data(path: str, lang: str = "en") -> InputData:
    filename: str = ".".join(os.path.split(path)[-1].split(".")[:-1])
    filetype: str = path.split(".")[-1].lower()
    # load data based on filetype
    # get loading object
    loader: BaseLoader = DOCUMENT_LOADERS.get(filetype, None)
    if not loader:
        raise Exception(f"File Type {filetype} was not implemented yet.")
    content_data: List[ContentItem] = loader.load_document(path)
    # create report object
    report_data: ContentDocument = ContentDocument(
        name=filename,
        taxonomy_schema="",
        lang=lang,
        xhtml=True,
        priority=0,
        namespaces={},
        content=content_data
    )
    return InputData(taxonomy=None, reports=[report_data])


import tempfile
import uuid
import os
import shutil
import logging

logger = logging.getLogger(__name__)

def create_tmp_dir() -> str:
    tmp_folder_path: str = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
    logger.info(f"Creating temporary directory at {tmp_folder_path}")
    os.mkdir(tmp_folder_path)
    return tmp_folder_path

def remove_tmp_dir(dir: str) -> None:
    shutil.rmtree(dir)
    logger.info(f"Removed temporary directory at {dir}")


from dataclasses import dataclass
from typing import Dict, List

from .ContentDataclasses import ContentDocument
from .TaxonomyDataclasses import TaxonomyDocument

@dataclass
class InputData:
    taxonomy: TaxonomyDocument
    reports: List[ContentDocument]

    @classmethod
    def from_dict(cls, data: dict) -> 'InputData':
        return cls(
            taxonomy=TaxonomyDocument.from_dict(data.get("taxonomy")) if "taxonomy" in data and data.get("taxonomy") else None,
            reports=[ContentDocument.from_dict(report) for report in data.get("reports", [])]
        )
    
    @classmethod
    def combine(cls, input_data_list: List["InputData"]) -> 'InputData':
        # combine taxonomies and update reports if necessary
        available_taxonomy_input_data: List[InputData] = [data for data in input_data_list if data.taxonomy]
        available_taxonomy_input_data.sort(key=lambda data: data.taxonomy.priority, reverse=True)
        target_taxonomy: TaxonomyDocument = None
        if available_taxonomy_input_data:
            target_taxonomy = available_taxonomy_input_data[0].taxonomy
            available_taxonomy_input_data = available_taxonomy_input_data[1:]
        for additional_data in available_taxonomy_input_data:
            update_element_map = target_taxonomy.add_taxonomy(additional_data.taxonomy)
            # update elements that were renamed in content documents
            for content_document in additional_data.reports:
                for content_item in content_document.content:
                    content_item.update_tags_elements(update_element_map)
        # sort and combine reports
        sorted_reports: Dict[str, List[ContentDocument]] = {}
        for input_data in input_data_list:
            for report in input_data.reports:
                if not report.name in sorted_reports:
                    sorted_reports[report.name] = []
                sorted_reports[report.name].append(report)
        combined_documents: List[ContentDocument] = []
        for report_name in sorted_reports:
            combined_documents.append(ContentDocument.combine(sorted_reports[report_name]))
        return cls(
            taxonomy=target_taxonomy,
            reports=combined_documents
        )
        

    def to_dict(cls) -> dict:
        return {
            "taxonomy": cls.taxonomy.to_dict() if cls.taxonomy else None,
            "reports": [report.to_dict() for report in cls.reports]
        }


from typing import Dict, Tuple
from lxml import etree
import logging
import math

from .ElementRender import render_content

from .PackageDataclasses import File
from .ContentDataclasses import AppliedTagTree, ContentDocument, ContentItem
from .BaseProducer import BaseProducer, INSTANCE_NAMESPACE, LINK_NAMESPACE, XLINK_NAMESPACE, XML_NAMESPACE, XSI_NAMESPACE, DIMENSIONS_NAMESPACE
from .utils import xml_to_string

logger = logging.getLogger(__name__)

class XbrlProducer(BaseProducer):

    def __init__(cls, document: ContentDocument, local_namespace: str = None, local_namespace_prefix: str = None, local_taxonomy_schema: str = None):
        super().__init__(
            document=document,
            local_namespace=local_namespace,
            local_namespace_prefix=local_namespace_prefix,
            local_taxonomy_schema=local_taxonomy_schema
        )

    def create_xbrl(cls) -> File:
        # Populate Namespaces
        namespace_map = {
            None: INSTANCE_NAMESPACE,
            "xml": XML_NAMESPACE,
            "link": LINK_NAMESPACE,
            "xlink": XLINK_NAMESPACE,
            "xsi": XSI_NAMESPACE,
            "xbrldi": DIMENSIONS_NAMESPACE
        }
        if cls.local_namespace:
            namespace_map[cls.local_namespace_prefix] = cls.local_namespace
        for namespace, prefix in cls.content_document.namespaces.items():
            namespace_map[prefix] = namespace

        # create base xbrl element
        root_element: etree._Element = etree.Element(
            f"{{{INSTANCE_NAMESPACE}}}xbrl",
            nsmap=namespace_map
        )

        #TODO: xsi:schemaLocation="http://mycompany.com/xbrl/taxonomy 102-01-SpecExample.xsd"

        # add schema ref
        cls.schema_url = cls.content_document.taxonomy_schema if cls.content_document.taxonomy_schema else cls.local_taxonomy_schema
        schema_ref: etree._Element = etree.SubElement(
            root_element, 
            f"{{{LINK_NAMESPACE}}}schemaRef",
            {
                f"{{{XLINK_NAMESPACE}}}href": cls.schema_url,
                f"{{{XLINK_NAMESPACE}}}type": "simple"
            }
        )
              
        # add contexts to header
        cls._add_context_elements(root_element)
        
        # Add Units
        cls._add_unit_elements(root_element)
        
        # Add Facts
        for content in cls.content_document.content:
            cls._convert_element(content, root_element)

        return File(name=f"{cls.content_document.name}.xbrl", content=xml_to_string(root_element))
    
    def _convert_element(cls, content_item: ContentItem, root_element: etree.Element) -> None:
        # check tags on structure
        tags = [tag for tag in content_item.tags if not tag.end_index and not tag.start_index]
        # create xbrl tag
        for tag in tags:
            applicable_namespace: str = cls.local_namespace
            if tag.namespace:
                applicable_namespace = tag.namespace
            # numeric fact
            if tag.attributes.unit_ref:
                new_element: etree._Element = etree.SubElement(
                    root_element,
                    f"{{{applicable_namespace}}}{tag.name}",
                    {
                        "contextRef": tag.context_id,
                        "unitRef": tag.attributes.unit_ref,
                        "decimals": str(tag.attributes.decimals)
                    }
                )
                if tag.attributes.nil:
                    new_element.attrib[f"{{{XSI_NAMESPACE}}}nil"] = "true"
                tmp_element: etree._Element = etree.Element("tmp")
                cls._rec_render_content(content_item, tmp_element)
                value = float(tmp_element.text) * (10**tag.attributes.scale)
                new_element.text = str(value)
            # otherwise it must be non-numeric
            else:
                # if it is not a enum
                if not tag.attributes.enumeration_values:
                    # get previous is if known
                    tag_id_base = f"{applicable_namespace}_{tag.name}_{tag.context_id}"
                    tag_element: etree._Element = cls.tag_id_tracker.get(tag_id_base, None)
                    if tag_element == None:
                        tag_element: etree._Element = etree.SubElement(
                            root_element,
                            f"{{{applicable_namespace}}}{tag.name}",
                            {
                                "contextRef": tag.context_id
                            }
                        )
                        cls.tag_id_tracker[tag_id_base] = tag_element
                    if tag.attributes.nil:
                        tag_element.attrib[f"{{{XSI_NAMESPACE}}}nil"] = "true"
                    cls._rec_render_content(
                        content_item,
                        tag_element
                    )
                # enumerations value
                else:
                    enum_tag_element: etree._Element = etree.SubElement(
                        root_element,
                        f"{{{applicable_namespace}}}{tag.name}",
                        {
                            "contextRef": tag.context_id
                        }
                    )
                    enum_tag_element.text = " ".join([enum.value(cls.local_namespace) for enum in tag.attributes.enumeration_values])
    
    def _rec_render_content(cls, content_item: ContentItem, tag_element: etree.Element) -> None:
        render_content(
            content_item,
            tag_element,
            cls._add_text,
            cls._rec_render_content
        )
                
    def _add_text(cls, element: etree.Element, tag_tree: AppliedTagTree, content: str) -> None:
        start_text: str = content[tag_tree.item.start_index:tag_tree.item.end_index]
        if tag_tree.children:
            start_text = content[tag_tree.item.start_index:tag_tree.children[0].item.start_index]
        element.text = start_text
        for child_index, child_tree in enumerate(tag_tree.children):
            current_element, child_element = cls._create_ixbrl_tag(child_tree.item, element)
            cls._add_text(child_element, child_tree, content)
            # if it is not the last element
            if child_index < len(tag_tree.children) - 1:
                child_element.tail = content[child_tree.item.end_index:tag_tree.children[child_index + 1].item.start_index]
            else:
                child_element.tail = content[child_tree.item.end_index:tag_tree.item.end_index]



from dataclasses import dataclass
from lxml import etree

class reversor:
    def __init__(self, obj):
        self.obj = obj

    def __eq__(self, other):
        return other.obj == self.obj

    def __lt__(self, other):
        return other.obj < self.obj

def xml_to_string(root: etree.Element, doctype: str = None) -> str:
    if not doctype:
        return etree.tostring(root, encoding="utf-8", xml_declaration=True, pretty_print=True).decode("utf-8")
    else:
        return etree.tostring(root, encoding="utf-8", doctype=doctype, xml_declaration=True, pretty_print=True).decode("utf-8")


from dataclasses import dataclass
import logging
from typing import Dict, List, Set, Tuple

from .PackageDataclasses import Tag
from .utils import reversor

logger = logging.getLogger(__name__)

@dataclass
class ContentDocument:
    name: str
    taxonomy_schema: str
    lang: str
    xhtml: bool
    priority: int
    namespaces: Dict[str, str]
    content: List['ContentItem']
    
    @classmethod
    def from_dict(cls, data: dict) -> 'ContentDocument':
        return cls(
            name=data.get("name"),
            taxonomy_schema=data.get("taxonomy_schema"),
            lang=data.get("lang"),
            xhtml=data.get("xhtml", True),
            priority=data.get("priority", 100),
            namespaces=data.get("namespaces"),
            content=[ContentItem.from_dict(item_data) for item_data in data.get("content")]
        )

    @classmethod
    def combine(cls, document_list: List["ContentDocument"]) -> "ContentDocument":
        # sort list by priority
        document_list.sort(key=lambda doc: doc.priority, reverse=True)
        # calculate base attributes if not net by document with highest priority
        taxonomy_schemas: List[str] = [doc.taxonomy_schema for doc in document_list]
        taxonomy_schema: str = taxonomy_schemas[0]
        if None in taxonomy_schemas:
            taxonomy_schema = None
        namespaces: Dict[str, str] = {}
        for doc_namespaces in [doc.namespaces for doc in document_list]:
            if doc_namespaces:
                namespaces.update(doc_namespaces)
        # merge content of each document
        content: List['ContentItem'] = []
        for document in document_list:
            for content_item in document.content:
                content.append(content_item)
        return cls(
            name=document_list[0].name,
            taxonomy_schema=taxonomy_schema,
            lang=document_list[0].lang,
            xhtml=any([doc.xhtml for doc in document_list]),
            priority=document_list[0].priority,
            namespaces=namespaces,
            content=content
        )
    
    def _get_applied_tags(cls) -> List['AppliedTag']:
        tags: List['AppliedTag'] = []
        for content_element in cls.content:       
            tags += content_element._get_applied_tags()
        return tags

    def to_dict(cls) -> dict:
        return {
            "name": cls.name,
            "taxonomy_schema": cls.taxonomy_schema,
            "lang": cls.lang,
            "xhtml": cls.xhtml,
            "priority": cls.priority,
            "namespaces": cls.namespaces,
            "content": [content_item.to_dict() for content_item in cls.content],
        }

@dataclass
class DocumentContext:
    entity: str
    entity_scheme: str
    end_date: str
    start_date: str
    dimensions: List['DocumentDimension']
    
    def euqals(cls, compare_context: "DocumentContext") -> bool:
        if cls.entity != compare_context.entity:
            return False
        if cls.entity_scheme != compare_context.entity_scheme:
            return False
        if cls.end_date != compare_context.end_date:
            return False
        if cls.start_date != compare_context.start_date:
            return False
        if not DocumentDimension.equal_dimensions(cls.dimensions, compare_context.dimensions):
            return False
        return True

@dataclass
class DocumentDimension:
    axis: 'Tag'
    member: 'Tag'
    typed_member_value: str

    @classmethod
    def from_dict(cls, data: dict) -> 'DocumentDimension':
        return cls(
            axis=Tag.from_dict(data.get("axis", {})),
            member=Tag.from_dict(data.get("member", {})),
            typed_member_value=data.get("typed_member_value", None)
        )
    
    def copy(cls) -> 'DocumentDimension':
        return cls.__class__(
            axis=cls.axis.copy(),
            member=cls.member.copy(),
            typed_member_value=cls.typed_member_value
        )
    
    def to_dict(cls) -> dict:
        return {
            "axis": cls.axis.to_dict(),
            "member": cls.member.to_dict(),
            "typed_member_value": cls.typed_member_value
        }

    @staticmethod
    def equal_dimensions(dimensions_a: List["DocumentDimension"], dimensions_b: List["DocumentDimension"]) -> bool:
        axis_unames_a: Dict[str, DocumentDimension] = {dim.axis.to_uname():dim for dim in dimensions_a}
        axis_unames_b: Dict[str, DocumentDimension] = {dim.axis.to_uname():dim for dim in dimensions_b}
        if set(axis_unames_a.keys()) != set(axis_unames_b.keys()):
            return False
        for axis in axis_unames_a.keys():
            if axis_unames_a[axis].typed_member_value != axis_unames_b[axis].typed_member_value:
                return False
            if axis_unames_a[axis].member.to_uname() != axis_unames_b[axis].member.to_uname():
                return False
        return True

@dataclass
class DocumentUnit:
    numerator: 'Tag'
    denominator: 'Tag'
    unit_id: str = None

    @classmethod
    def from_dict(cls, data: dict) -> 'DocumentUnit':
        return cls(
            numerator=Tag.from_dict(data.get("numerator", {})),
            denominator=Tag.from_dict(data.get("denominator", {})) if data.get("denominator", False) else None
        )
    
    def copy(cls) -> 'DocumentUnit':
        return cls.__class__(
            numerator=cls.numerator.copy(),
            denominator=cls.denominator.copy() if cls.denominator else None,
            unit_id=cls.unit_id
        )

    def equals(cls, compare_unit: "DocumentUnit") -> bool:
        # compare numerator
        numerator_a: str = None
        if cls.numerator:
            numerator_a = cls.numerator.to_uname()
        numerator_b: str = None
        if compare_unit.numerator:
            numerator_b = compare_unit.numerator.to_uname()
        if numerator_a != numerator_b:
            return False
        # compare denonimator
        denominator_a: str = None
        if cls.denominator:
            denominator_a = cls.denominator.to_uname()
        denominator_b: str = None
        if compare_unit.denominator:
            denominator_b = compare_unit.denominator.to_uname()
        if denominator_a != denominator_b:
            return False
        return True
    
    def to_dict(cls) -> dict:
        return {
            "numerator": cls.numerator.to_dict(),
            "denominator": cls.denominator.to_dict() if cls.denominator else None
        }

@dataclass
class EnumerationValue(Tag):

    @classmethod
    def from_dict(cls, data: dict) -> 'EnumerationValue':
        return cls(
            name = data.get("name"),
            namespace = data.get("namespace")
        )
    
    def value(cls, extension_namespace: str) -> str:
        if cls.namespace:
            return f"{cls.namespace}#{cls.name}"
        return f"{extension_namespace}#{cls.name}"

    def copy(cls) -> 'EnumerationValue':
        return cls.__class__(
            name=cls.name,
            namespace=cls.namespace
        )

    def to_dict(cls) -> dict:
        return {
            "name": cls.name,
            "namespace": cls.namespace
        }

@dataclass
class TagAttributes:
    # only non-numeric
    escape: bool
    enumeration_values: List['EnumerationValue']
    # both
    format: Tag
    nil: bool
    # only numeric
    decimals: int
    scale: int
    unit: DocumentUnit
    sign: bool
    # will be filled automatically
    unit_ref: str = None

    @classmethod
    def from_dict(cls, data: dict) -> 'TagAttributes':
        return cls(
            escape=data.get("escape", False),
            enumeration_values=[EnumerationValue.from_dict(d) for d in data.get("enumeration_values", [])],
            format=Tag.from_dict(data.get("format")) if "format" in data and data["format"] else None,
            nil=data.get("nil", False),
            decimals=data.get("decimals", 0),
            scale=data.get("scale", 0),
            unit=DocumentUnit.from_dict(data.get("unit")) if "unit" in data and data["unit"] else None,
            sign=data.get("sign", False)
        )
    
    def copy(cls) -> 'TagAttributes':
        return cls.__class__(
            escape=cls.escape,
            enumeration_values=[ev.copy() for ev in cls.enumeration_values],
            format=cls.format.copy() if cls.format else None,
            nil=cls.nil,
            decimals=cls.decimals,
            scale=cls.scale,
            unit=cls.unit.copy() if cls.unit else None,
            sign=cls.sign
        )
    
    def to_dict(cls) -> dict:
        return {
            "escape": cls.escape,
            "enumeration_values": [ev.to_dict() for ev in cls.enumeration_values],
            "format": cls.format.to_dict() if cls.format else None,
            "nil": cls.nil,
            "decimals": cls.decimals,
            "scale": cls.scale,
            "unit": cls.unit.to_dict() if cls.unit else None,
            "sign": cls.sign
        }

@dataclass
class AppliedTag(Tag):
    attributes: TagAttributes
    entity: str
    entity_scheme: str
    start_date: str
    end_date: str
    dimensions: List['DocumentDimension']
    start_index: int = None
    end_index: int = None
    # this is generated while the production of the file
    context_id: str = None

    @classmethod
    def from_dict(cls, data: dict) -> 'AppliedTag':
        return cls(
            namespace=data.get("namespace"),
            name=data.get("name"),
            attributes=TagAttributes.from_dict(data.get("attributes", {})),
            entity=data.get("entity"),
            entity_scheme=data.get("entity_scheme"),
            start_date=data.get("start_date"),
            end_date=data.get("end_date"),
            dimensions=[DocumentDimension.from_dict(dd) for dd in data.get("dimensions", [])],
            start_index=data.get("start_index", None),
            end_index=data.get("end_index", None)
        )
    
    @classmethod
    def empty(cls, start_index: int = 0, end_index: int = 0) -> 'AppliedTag':
        return cls(
            namespace="",
            name="",
            attributes={},
            entity="",
            entity_scheme="",
            start_date="",
            end_date="",
            dimensions=[],
            start_index=start_index,
            end_index=end_index
        )
    
    def to_document_context(cls) -> DocumentContext:
        return DocumentContext(
            entity=cls.entity,
            entity_scheme=cls.entity_scheme,
            end_date=cls.end_date,
            start_date=cls.start_date,
            dimensions=cls.dimensions
        )

    def contains_tag(cls, compare_tag: 'AppliedTag') -> bool:
        return cls.start_index <= compare_tag.start_index and compare_tag.end_index <= cls.end_index

    def find_intercept(cls, compare_tag: 'AppliedTag') -> int:
        #  |---------------------|         cls
        #        |--------------------|    compare tag
        #                        x this index
        if compare_tag.start_index < cls.end_index and cls.end_index < compare_tag.end_index:
            return cls.end_index
        return -1
    
    def split(cls, index: int) -> List['AppliedTag']:
        splitted_tag = cls.__class__(
                namespace=cls.namespace,
                name=cls.name,
                attributes=cls.attributes.copy(),
                entity=cls.entity,
                entity_scheme=cls.entity_scheme,
                start_date=cls.start_date,
                end_date=cls.end_date,
                dimensions=[dd.copy() for dd in cls.dimensions],
                start_index=index,
                end_index=cls.end_index,
                context_id=cls.context_id
            )
        cls.end_index = index
        return splitted_tag
        
    def to_dict(cls) -> dict:
        return {
            "namespace": cls.namespace,
            "name": cls.name,
            "attributes": cls.attributes.to_dict(),
            "entity": cls.entity,
            "entity_scheme": cls.entity_scheme,
            "start_date": cls.start_date,
            "end_date": cls.end_date,
            "dimensions": [dd.to_dict() for dd in cls.dimensions],
            "start_index": cls.start_index,
            "end_index": cls.end_index
        }
    
    @staticmethod
    def _sort(tags: List['AppliedTag']) -> List['AppliedTag']:
        return sorted(tags, key=lambda x: (x.start_index, reversor(x.end_index), x.attributes.unit != None))

    @staticmethod
    def create_tree(tags: List['AppliedTag'], content_len: int) -> 'AppliedTagTree':
        tags = AppliedTag._sort(tags)
        current_index: int = 0
        while current_index < len(tags):
            current_tag: AppliedTag = tags[current_index]
            new_tags: List[AppliedTag] = []
            for comparison_tag in tags[current_index + 1:]:
                intercept_index: int = current_tag.find_intercept(comparison_tag)
                if intercept_index != -1:
                    new_tags.append(comparison_tag.split(intercept_index))
            tags += new_tags
            tags = AppliedTag._sort(tags)
            current_index += 1
        tree_wrapper: AppliedTagTree = AppliedTagTree(
            AppliedTag.empty(end_index=content_len), 
            [], 
            True
        )
        for tag in tags:
            tree_wrapper.add_tag(tag)
        return tree_wrapper

@dataclass
class AppliedTagTree:
    item: AppliedTag
    children: List['AppliedTagTree']
    wrapper: bool = False

    def add_tag(cls, new_tag: AppliedTag) -> None:
        # check if it needs to be added to a subchild or granchild (recursive)
        for tag in cls.children:
            if tag.item.contains_tag(new_tag):
                tag.add_tag(new_tag)
                return
        # if not a subchild, add as new child to this one
        cls.children.append(AppliedTagTree(new_tag, []))

@dataclass(kw_only=True)
class ContentItem:
    type: str
    tags: List[AppliedTag]

    TYPE_TITLE: str = "TITLE"
    TYPE_PARAGRAPH: str = "PARAGRAPH"
    TYPE_TABLE: str = "TABLE"
    TYPE_IMAGE: str = "IMAGE"
    TYPE_LIST: str = "LIST"
    TYPE_BASE_XBRL: str = "BASE_XBRL"

    @classmethod
    def from_dict(cls, data: dict) -> 'ContentItem':
        match data.get("type"):
            case cls.TYPE_TITLE:
                return TitleItem.from_dict(data)
            case cls.TYPE_PARAGRAPH:
                return ParagraphItem.from_dict(data)
            case cls.TYPE_TABLE:
                return TableItem.from_dict(data)
            case cls.TYPE_IMAGE:
                return ImageItem.from_dict(data)
            case cls.TYPE_LIST:
                return ListItem.from_dict(data)
            case cls.TYPE_BASE_XBRL:
                return BaseXbrlItem.from_dict(data)
            case _:
                logger.error(f"Content Item Type '{data.get('type')}' is not implemented yet.")
                return cls(
                    data.get("type"), 
                    [AppliedTag.from_dict(tag_data) for tag_data in data.get("tags", [])]
                )

    def update_tags_elements(cls, element_update_map: Dict[str, str]) -> None:
        raise Exception(f"The function update_tags_elements was not implemented for the content type {cls.type}.")

    def _get_applied_tags(cls) -> List[AppliedTag]:
        raise Exception(f"The function _get_applied_tags was not implemented for the content type {cls.type}.")

    def to_dict(cls) -> dict:
        return {
            "type": cls.type,
            "tags": [tag.to_dict() for tag in cls.tags]
        }

@dataclass
class TitleItem(ContentItem):
    content: str
    level: int

    @classmethod
    def from_dict(cls, data: dict) -> 'TitleItem':
        return cls(
            type=data.get("type"),
            content=data.get("content"),
            level=data.get("level"),
            tags=[AppliedTag.from_dict(tag_data) for tag_data in data.get("tags", [])]
        )
    
    def update_tags_elements(cls, element_update_map: Dict[str, str]) -> None:
        for tag in cls.tags:
            if not tag.namespace and tag.name in element_update_map:
                tag.name = element_update_map[tag.name]

    def _get_applied_tags(cls) -> List[AppliedTag]:
        return cls.tags

    def to_dict(cls) -> dict:
        return {
            "type": cls.type,
            "content": cls.content,
            "level": cls.level,
            "tags": [tag.to_dict() for tag in cls.tags]
        }
    
@dataclass
class ParagraphItem(ContentItem):
    content: str

    @classmethod
    def from_dict(cls, data: dict) -> 'TitleItem':
        return cls(
            type=data.get("type"),
            content=data.get("content"),
            tags=[AppliedTag.from_dict(tag_data) for tag_data in data.get("tags", [])]
        )

    def update_tags_elements(cls, element_update_map: Dict[str, str]) -> None:
        for tag in cls.tags:
            if not tag.namespace and tag.name in element_update_map:
                tag.name = element_update_map[tag.name]

    def _get_applied_tags(cls) -> List[AppliedTag]:
        return cls.tags

    def to_dict(cls) -> dict:
        return {
            "type": cls.type,
            "content": cls.content,
            "tags": [tag.to_dict() for tag in cls.tags]
        }
    
@dataclass
class TableItem(ContentItem):
    rows: List['TableRow']

    @classmethod
    def from_dict(cls, data: dict) -> 'TitleItem':
        return cls(
            type=data.get("type"),
            rows=[TableRow.from_dict(row_data) for row_data in data.get("rows", [])],
            tags=[AppliedTag.from_dict(tag_data) for tag_data in data.get("tags", [])]
        )

    def update_tags_elements(cls, element_update_map: Dict[str, str]) -> None:
        for tag in cls.tags:
            if not tag.namespace and tag.name in element_update_map:
                tag.name = element_update_map[tag.name]
        for row in cls.rows:
            row.update_tags_elements(element_update_map)

    def _get_applied_tags(cls) -> List[AppliedTag]:
        tags: List[AppliedTag] = cls.tags
        for row in cls.rows:
            tags += row._get_applied_tags()
        return tags
    
    def to_dict(cls) -> dict:
        return {
            "type": cls.type,
            "rows": [row.to_dict() for row in cls.rows],
            "tags": [tag.to_dict() for tag in cls.tags]
        }

@dataclass
class TableRow:
    cells: List['TableCell']

    @classmethod
    def from_dict(cls, data: dict) -> 'TableRow':
        return cls(
            cells=[TableCell.from_dict(cell_data) for cell_data in data.get("cells", [])]
        )
    
    def update_tags_elements(cls, element_update_map: Dict[str, str]) -> None:
        for cell in cls.cells:
            cell.update_tags_elements(element_update_map)

    def _get_applied_tags(cls) -> List[AppliedTag]:
        tags: List[AppliedTag] = []
        for cell in cls.cells:
            tags += cell._get_applied_tags()
        return tags

    def to_dict(cls) -> dict:
        return {
            "cells": [cell.to_dict() for cell in cls.cells]
        }

@dataclass
class TableCell:
    content: List[ContentItem]
    header: bool
    rowspan: int
    colspan: int

    @classmethod
    def from_dict(cls, data: dict) -> 'TableCell':
        return cls(
            content=[ContentItem.from_dict(content_data) for content_data in data.get("content", [])],
            header=data.get("header", False),
            rowspan=data.get("rowspan", 1),
            colspan=data.get("colspan", 1)
        )
    
    def update_tags_elements(cls, element_update_map: Dict[str, str]) -> None:
        for content_item in cls.content:
            content_item.update_tags_elements(element_update_map)

    def _get_applied_tags(cls) -> List[AppliedTag]:
        tags: List[AppliedTag] = []
        for content_element in cls.content:
            tags += content_element._get_applied_tags()
        return tags
    
    def to_dict(cls) -> dict:
        return {
            "content": [item.to_dict() for item in cls.content],
            "header": cls.header,
            "rowspan": cls.rowspan,
            "colspan": cls.colspan
        }


@dataclass
class ImageItem(ContentItem):
    image_data: str

    @classmethod
    def from_dict(cls, data: dict) -> 'TitleItem':
        return cls(
            type=data.get("type"),
            image_data=data.get("image_data"),
            tags=[AppliedTag.from_dict(tag_data) for tag_data in data.get("tags", [])]
        )
    
    def update_tags_elements(cls, element_update_map: Dict[str, str]) -> None:
        for tag in cls.tags:
            if not tag.namespace and tag.name in element_update_map:
                tag.name = element_update_map[tag.name]

    def _get_applied_tags(cls) -> List[AppliedTag]:
        return cls.tags

    def to_dict(cls) -> dict:
        return {
            "type": cls.type,
            "image_data": cls.image_data,
            "tags": [tag.to_dict() for tag in cls.tags]
        }
    
@dataclass
class ListItem(ContentItem):
    elements: List['ListElement']
    ordered: bool

    @classmethod
    def from_dict(cls, data: dict) -> 'ListItem':
        return cls(
            type=data.get("type"),
            elements=[ListElement.from_dict(element_data) for element_data in data.get("elements", [])],
            ordered=data.get("ordered", False),
            tags=[AppliedTag.from_dict(tag_data) for tag_data in data.get("tags", [])]
        )

    def update_tags_elements(cls, element_update_map: Dict[str, str]) -> None:
        for tag in cls.tags:
            if not tag.namespace and tag.name in element_update_map:
                tag.name = element_update_map[tag.name]
        for element in cls.elements:
            element.update_tags_elements(element_update_map)

    def _get_applied_tags(cls) -> List[AppliedTag]:
        tags: List[AppliedTag] = cls.tags
        for list_element in cls.elements:
            tags += list_element._get_applied_tags()
        return tags
    
    def to_dict(cls) -> dict:
        return {
            "type": cls.type,
            "elements": [element.to_dict() for element in cls.elements],
            "ordered": cls.ordered,
            "tags": [tag.to_dict() for tag in cls.tags]
        }
    
@dataclass
class ListElement:
    content: List[ContentItem]

    @classmethod
    def from_dict(cls, data: dict) -> 'ListElement':
        return cls(
            content=[ContentItem.from_dict(element_content) for element_content in data.get("content", [])]
        )

    def update_tags_elements(cls, element_update_map: Dict[str, str]) -> None:
        for content_item in cls.content:
            content_item.update_tags_elements(element_update_map)

    def _get_applied_tags(cls) -> List[AppliedTag]:
        tags: List[AppliedTag] = []
        for content_element in cls.content:
            tags += content_element._get_applied_tags()
        return tags

    def to_dict(cls) -> dict:
        return {
            "content": [content_item.to_dict() for content_item in cls.content]
        }
    
@dataclass
class BaseXbrlItem(ContentItem):
    content: str

    @classmethod
    def from_dict(cls, data: dict) -> 'BaseXbrlItem':
        return cls(
            type=data.get("type"),
            content=data.get("content"),
            tags=[AppliedTag.from_dict(tag_data) for tag_data in data.get("tags", [])]
        )

    def update_tags_elements(cls, element_update_map: Dict[str, str]) -> None:
        for tag in cls.tags:
            if not tag.namespace and tag.name in element_update_map:
                tag.name = element_update_map[tag.name]
        
    def _get_applied_tags(cls) -> List[AppliedTag]:
        return cls.tags
    
    def to_dict(cls) -> dict:
        return {
            "type": cls.type,
            "content": cls.content,
            "tags": [tag.to_dict() for tag in cls.tags]
        }


import os
import shutil
import logging
import json
from typing import List
from fastapi import FastAPI, UploadFile, File, Form, BackgroundTasks
from fastapi.responses import FileResponse

from ..utils.tmp_dirs import create_tmp_dir, remove_tmp_dir
from ..__init__ import convert_document, validate_input_data, load_input_data, create_xbrl, InputData

logger = logging.getLogger(__name__)

def save_file(tmp_dir: str, file: UploadFile) -> str:
    filepath: str = os.path.join(tmp_dir, file.filename)
    logger.info(f"Saving uploaded file to {filepath}")
    with open(filepath, "wb") as f:
        shutil.copyfileobj(file.file, f)
    return filepath

app = FastAPI()

@app.get("/")
async def root():
    return {"xbrl-forge": "Hello There!"}

@app.post("/convert_document")
async def convert_document_endpoint(document: UploadFile = File(...)):
    try:
        tmp_dir: str = create_tmp_dir()
        document_filepath: str = save_file(tmp_dir, document)
        return convert_document(document_filepath).to_dict()
    finally:
        remove_tmp_dir(tmp_dir)

@app.post("/validate_input_data")
async def validate_input_data_endpoint(input_data: str = Form(...)):
    valdation_result: dict = {
        "valid": True,
        "message": "The input data is valid against the JSON Schema"
    }
    try:
        input_data_dict: dict = json.loads(input_data)
        validate_input_data(input_data_dict)
    except Exception as e:
        valdation_result["valid"] = False
        valdation_result["message"] = str(e)
    return valdation_result

@app.post("/create_xbrl")
async def create_xbrl_endpoint(bg_tasks: BackgroundTasks, input_data_list: str = Form(...), xthml_template: str = Form(None)):
    loaded_data_dicts: List[dict] = json.loads(input_data_list)
    loaded_data: List[InputData] = [load_input_data(input_data_dict) for input_data_dict in loaded_data_dicts]
    result = create_xbrl(loaded_data, xthml_template)
    try:
        tmp_dir: str = create_tmp_dir()
        bg_tasks.add_task(remove_tmp_dir, tmp_dir)
        file_path: str = result.create_package(tmp_dir)
        return FileResponse(file_path, background=bg_tasks, headers={'Content-Disposition': f'attachment; filename={file_path.split("/")[-1]}'})
    except Exception as e:
        logger.error(str(e))
        remove_tmp_dir(tmp_dir)
    